# -*- coding: utf-8 -*-

# module functions

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
A number of mathematical functions are defined here that can be used with
gummys.
"""

import numpy as np
from .gummy import gummy,jummy
from .ummy import ummy
from numbers import Complex
from .dfunc import _call

def _callg(f,*args):
    # decide whether to call jummy.apply, gummy.apply or ummy.apply
    g = None
    u = None
    c = False
    for a in args:
        if isinstance(a,jummy):
            return a.__array_ufunc__(f,'__call__',*args)
        if isinstance(a,gummy):
            g = a
        if isinstance(a,ummy):
            u = a
        if isinstance(a,Complex):
            c = True
    if g is not None:
        if c:
            return jummy(g).__array_ufunc__(f,'__call__',*args)
        return g.__array_ufunc__(f,'__call__',*args)
    if u is not None:
        return u.__array_ufunc__(f,'__call__',*args)
    return _call(f,*args)

def _bcallg(f,*args):
    # broadcast f across any arrays in args.
    bargs = np.broadcast(*args)
    if bargs.shape == ():
        return _callg(f,*args)
    ret = np.array([_callg(f,*a) for a in bargs])
    ret = ret.reshape(bargs.shape)
    return ret


def sin(x):
    """
    Returns the sine of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.sin,x)
    
def cos(x):
    """
    Returns the cosine of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.cos,x)

def tan(x):
    """
    Returns the tangent of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.tan,x)
 
def arcsin(x):
    """
    Returns the inverse sine of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.arcsin,x)

def arccos(x):
    """
    Returns the inverse cosine of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.arccos,x)

def arctan(x):
    """
    Returns the inverse tangent of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.arctan,x)

def arctan2(x1,x2):
    """
    Returns the inverse tangent of x1/x2 choosing the quadrant correctly where 
    x1 and x2 can be float, complex, gummy or jummy.  
    """
    return _bcallg(np.arctan2,x1,x2)
    
def sinh(x):
    """
    Returns the hyperbolic sine of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.sinh,x)

def cosh(x):
    """
    Returns the hyperbolic cosine of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.cosh,x)

def tanh(x):
    """
    Returns the hyperbolic tangent of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.tanh,x)

def arcsinh(x):
    """
    Returns the inverse hyperbolic sine of x where x can be float, complex, gummy 
    or jummy.
    """
    return _bcallg(np.arcsinh,x)

def arccosh(x):
    """
    Returns the inverse hyperbolic cosine of x where x can be float, complex, gummy 
    or jummy.
    """
    return _bcallg(np.arccosh,x)

def arctanh(x):
    """
    Returns the inverse hyperbolic tangent of x where x can be float, complex, 
    gummy or jummy.
    """
    return _bcallg(np.arctanh,x)

def exp(x):
    """
    Returns e to the power of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.exp,x)

def exp2(x):
    """
    Returns 2 to the power of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.exp2,x)

def expm1(x):
    """
    Returns exp(x) - 1 where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.expm1,x)

def log(x):
    """
    Returns the natural logrithm of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.log,x)
    
def log2(x):
    """
    Returns the log base 2 of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.log2,x)

def log10(x):
    """
    Returns the log base 10 of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.log10,x)

def log1p(x):
    """
    Returns the natural logrithm of x plus 1 where x can be float, complex, gummy 
    or jummy.
    """
    return _bcallg(np.log1p,x)

def logaddexp(x1,x2):
    """
    Returns the log(exp(x1) + exp(x2)) where x1 and x2 can be float, complex, gummy 
    or jummy and log is the natural logrithm.
    """
    return _bcallg(np.logaddexp,x1,x2)

def logaddexp2(x1,x2):
    """
    Returns the log2(2**x1 + 2**x2) where x1 and x2 can be float, complex, gummy 
    or jummy and log2 is the logrithm to base 2.
    """
    return _bcallg(np.logaddexp2,x1,x2)

def sqrt(x):
    """
    Returns the square root of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.sqrt,x)

def cbrt(x):
    """
    Returns the cube root of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.cbrt,x)

def square(x):
    """
    Returns the square of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.square,x) 

def add(x1,x2):
    """
    returns x1 + x2
    """
    return _bcallg(np.add,x1,x2)

def subtract(x1,x2):
    """
    returns x1 - x2
    """
    return _bcallg(np.subtract,x1,x2)

def negative(x):
    """
    returns -x
    """
    return _bcallg(np.negative,x)

def multiply(x1,x2):
    """
    returns x1 * x2
    """
    return _bcallg(np.multiply,x1,x2)

def divide(x1,x2):
    """
    returns x1 / x2
    """
    return _bcallg(np.divide,x1,x2)

def true_divide(x1,x2):
    """returns x1 / x2
    """
    return _bcallg(np.true_divide,x1,x2)

def floor_divide(x1,x2):
    """
    returns x1 // x2
    """
    return _bcallg(np.floor_divide,x1,x2)

def reciprocal(x):
    """
    returns 1/x
    """
    return _bcallg(np.reciprocal,x)

def power(x1,x2):
    """
    returns x1**x2
    """
    return _bcallg(np.power,x1,x2)

def absolute(x):
    """
    Returns the absoulte value of x where x can be float, complex, gummy or jummy.
    This is equivalent to abs(x).
    """
    return _bcallg(np.absolute,x)

def mod(x1,x2):
    """
    returns x1 % x2
    """
    return _bcallg(np.mod,x1,x2)

def remainder(x1,x2):
    """
    returns x1 % x2
    """
    return _bcallg(np.remainder,x1,x2)

def divmod(x1,x2):
    """
    returns (x1 // x2, x1 % x2)
    """
    return _bcallg(np.divmod,x1,x2)

def modf(x1,x2):
    """
    returns (x1 % 1, x1 // 1), a tuple of integer and fractional parts
    """
    return _bcallg(np.modf,x1,x2)

def angle(x):
    """
    Returns the complex argument of x, where x can be float, complex, gummy or 
    jummy and the return value is a float if x is float or complex and gummy
    if x is gummy or jummy.
    """
    return _bcallg(np.angle,x)

def real(x):
    """
    returns x.real
    """
    return _bcallg(np.real,x)

def imag(x):
    """
    returns x.imag
    """
    return _bcallg(np.imag,x)

def conj(x):
    """
    returns x.conjugate(), the complex conjugate of x
    """
    return _bcallg(np.conj,x)

def around(x,n=0):
    """
    Returns x rounded to n digits where x can be float, complex, gummy or jummy
    but n must be int.
    """
    return _bcallg(np.around,x,n)

def rint(x):
    """
    Returns x rounded to the nearest integer value where x can be float, complex, 
    gummy or jummy.
    """
    return _bcallg(np.rint,x)

def fix(x):
    """
    Returns x rounded towards zero where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.fix,x)

def floor(x):
    """
    Returns the floor of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.floor,x)

def ceil(x):
    """
    Returns the ceiling of x where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.ceil,x)

def trunc(x):
    """
    Returns x rounded towards zero where x can be float, complex, gummy or jummy.
    """
    return _bcallg(np.trunc,x)

def heaviside(x,h0):
    """
    Heavyside function of x, h0 is the value at x = 0
    """
    return _bcallg(np.heavyside,x,h0)

def sign(x):
    """
    sign of x
    """
    return _bcallg(np.sign,x)

def sum(*args,**kwds):
    """
    Alias for numpy.sum
    """
    return np.sum(*args,**kwds)

def prod(*args,**kwds):
    """
    Alias for numpy.prod
    """
    return np.prod(*args,**kwds)

def cumsum(*args,**kwds):
    """
    Alias for numpy.cumsum
    """
    return np.cumsum(*args,**kwds)

def cumprod(*args,**kwds):
    """
    Alias for numpy.cumprod
    """
    return np.cumprod(*args,**kwds)

def diff(*args,**kwds):
    """
    Alias for numpy.diff
    """
    return np.diff(*args,**kwds)

def ediff1d(*args,**kwds):
    """
    Alias for numpy.ediff1d
    """
    return np.ediff1d(*args,**kwds)

def gradient(*args,**kwds):
    """
    Alias for numpy.gradient
    """
    return np.gradient(*args,**kwds)

def cross(*args,**kwds):
    """
    Alias for numpy.cross
    """
    return np.cross(*args,**kwds)
    




# -*- coding: utf-8 -*-

# module relunits

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module is loaded by the gummy.units module and is not intended be be
imported directly.  Dimensionless units are defined here.
"""
from .unit import _BuiltinLib,Unit,Conversion,one
from .ummy import MFraction

class RatioUnit(Unit):
    """RatioUnit is used for dimensionless units like % where powers, e.g. %**2,
    are not desired.
    """
    def _cpow(self,v):
        if v == 1 or v == -1:
            return ([[self,v]],1)
        c = self.convert(1,one)
        if int(v) == v:
            if v > 0:
                return ([[self,1]],c**(v - 1))
            else:
                return ([[self,1]],c**(v + 1))
        return ([[one,1]],c**v)

with _BuiltinLib():
    # The \t at the beggining of the symbol causes the space between the numerical
    # value and the unit symbol per the Chicago Manual of Style (but not the 
    # SI brochure).
    RatioUnit('percent','\t%',Conversion(one,MFraction('0.01')),short_name='%',
         latex_symbol='\t\\%',add_symbol=False)
    
    RatioUnit('part per million','ppm',Conversion(one,MFraction('1e-6')),add_symbol=True)
    RatioUnit('part per billion','ppb',Conversion(one,MFraction('1e-9')),add_symbol=True)
    RatioUnit('part per trillion','ppt',Conversion(one,MFraction('1e-12')),add_symbol=True)
    RatioUnit('part per quadrillion','ppq',Conversion(one,MFraction('1e-15')),add_symbol=True)



# -*- coding: utf-8 -*-

# module logunit

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
Unit and Conversion sub-classes are defined here to implement logarithmic 
units.
"""

import numpy as np
from .gummy import gummy
from .unit import one,_CompositeUnit
from .nonlinearunit import NonlinearConversion,NonlinearUnit
from .exceptions import IncompatibleUnitsError

class LogConversion(NonlinearConversion):   
    def __init__(self,reference,multiplier,log_base,log_func,offset=0):
        if isinstance(reference,gummy):
            self._unit = reference._unit
            self._rf = gummy(reference.x,reference.u)
            self._rf._ref = reference._ref
            self._rf._refs = reference._refs
        else:
            self._unit = one
            self._rf = reference
    
        self.reference = reference
        
        self.multiplier = multiplier           
        self.log_base = log_base
        self.log_func = lambda x: log_func(x) if x > 0 else -float('inf')
        self._lnbase = np.log(float(log_base))
        self.offset = offset
        
    def _to(self,g):
        g = (g - self.offset)/self.multiplier
        def f(x):
            if x == -float('inf'):
                return 0.0
            try:
                return self.log_base**x
            except OverflowError:
                return float('inf')
        def d(x):
            try:
                return self._lnbase*self.log_base**x
            except OverflowError:
                return 1
        ret = self._rf*gummy.apply(f,d,g)
        return ret
        
        
    def _frm(self,g):
        g = g/self._rf
        def f(x):
            return self.log_func(x)
        def d(x):
            if x != 0:
                return 1/(x*self._lnbase)
            return 1
        ret = self.multiplier*gummy.apply(f,d,g) + self.offset
        return ret
        
    def copy(self):
        r = LogConversion(self.reference,self.multiplier,self.log_base,self.log_func)
        r.parent = self.parent
        return r
    
    
class LogUnit(NonlinearUnit):   
    def __init__(self,*p,**kwds):
        super().__init__(*p,**kwds)
        self.reference = self.conversion.reference
        if isinstance(self.reference,gummy):
            self.referencef = self.reference.x
        else:
            self.referencef = self.reference
        self.multiplier = self.conversion.multiplier
        self.log_base = self.conversion.log_base
        
    def get_composite(self,ul):
        n = 0
        for u,e in ul:
            if not u.linear:
                n += 1
        if n > 1:
            raise IncompatibleUnitsError('unit ' + self.tostring() + ' may not be combined with other nonlinear units')
            
        return _LogCompositeUnit
        
    def zero(self):
        return -float('inf')
    
    def _add(self,a,bunit,b,aconv):
        if bunit is not self:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring() + '; automatic conversion is disabled with LogUnit instances')
        return (self._nummy_add(a,b),self)
    
    def _radd(self,a,bunit,b,aconv):
        if bunit is not self:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring() + '; automatic conversion is disabled with LogUnit instances')
        return (self._nummy_radd(a,b),self)
    
    def _sub(self,a,bunit,b,aconv):
        if bunit is not self:
            raise IncompatibleUnitsError('a quantity with unit ' + bunit.tostring() + ' may not be subtracted from a quantity with unit ' + self.tostring() + '; automatic conversion is disabled with LogUnit instances')
        return (self._nummy_sub(a,b),self)
    
    def _rsub(self,a,bunit,b,aconv):
        if bunit is not self:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be subtracted from a quantity with unit ' + bunit.tostring() + '; automatic conversion is disabled with LogUnit instances')
        return (self._nummy_rsub(a,b),self)
    
    def _mul(self,a,bunit,b,aconv):
        if not bunit.linear:
            raise IncompatibleUnitsError('only quantities with linear units may multiply or divide a quanity with unit ' + self.tostring())
        
        if aconv:
            un,c = self._mul_cancel(bunit)
        else:
            un,c = bunit._mul_cancel(self)
        
        return (self._nummy_mul(self._nummy_mul(a,b),c),un)
    
    def _rmul(self,a,bunit,b,aconv):
        if not bunit.linear:
            raise IncompatibleUnitsError('only quantities with linear units may multiply or divide a quanity with unit ' + self.tostring())
        
        if aconv:
            un,c = self._mul_cancel(bunit)
        else:
            un,c = bunit._mul_cancel(self)
        
        return (self._nummy_mul(self._nummy_rmul(a,b),c),un)
    
    def _truediv(self,a,bunit,b,aconv):
        if not bunit.linear:
            raise IncompatibleUnitsError('only quantities with linear units may multiply or divide a quanity with unit ' + self.tostring())
        
        if aconv:
            un,c = self._div_cancel(bunit)
        else:
            un,c = bunit._rdiv_cancel(self)
        
        return (self._nummy_mul(self._nummy_truediv(a,b),c),un)
    
    def _neg(self,a):
        return (self._nummy_neg(a),self)
    
    def _pos(self,a):
        return (self._nummy_pos(a),self)
    
    def _abs(self,a):
        return (self._nummy_abs(a),self)
        
    def __mul__(self,a):
        if not a.linear:
            raise IncompatibleUnitsError('only linear units may multiply or divide unit ' + self.tostring())
        return _CompositeUnit(self.units + a.units)
    
    def __truediv__(self,a):
        if not a.linear:
            raise IncompatibleUnitsError('only linear units may multiply or divide unit ' + self.tostring())
        vi = [(e[0],-e[1]) for e in a.units]
        return _CompositeUnit(self.units + vi)


class _LogCompositeUnit(LogUnit,_CompositeUnit):
    pass


# -*- coding: utf-8 -*-

# module siunits

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module is loaded by the gummy.units module and is not intended be be
imported directly.  

The most of the units here are from the SI Brochure, draft 9th edition published 
by the BIPM.
"""

from .gummy import gummy, _lg10
from .ummy import MFraction
from .unit import _BuiltinLib,Conversion,Unit
from .prefixedunit import PrefixedUnit,BinaryPrefixedUnit
from .logunit import LogUnit,LogConversion
from .offsetunit import OffsetUnit,OffsetConversion
from .functions import sqrt
import numpy as np

pi = np.pi
e = np.e

# constants from draft of the 9 edition of the SI brochure dated 18 Dec. 2018
_const_c = 299792458 # speed of light in m/s
_const_h = MFraction('6.62607015e-34') # planck constant in J s
_const_hbar = _const_h/(2*pi)
_const_e = MFraction('1.602176634e-19') # electron charge in C
_const_k = MFraction('1.380649e-23') # Boltzmann constant in J/K
_const_dalton = gummy(1.660539040e-27,0.000000020e-27) # unified atomic mass unit in kg

# constants from CODATA 2014
_const_G = gummy(6.67408e-11,0.00031e-11) # gravitational constant in m**3/kg s**2
_const_me = gummy(9.10938356e-31,0.00000011e-31) # mass of the electron in kg
_const_alpha = gummy(7.2973525664e-3,0.0000000017e-3) # fine structure constant
_const_rydberg = gummy(10973731.568508,0.000065) # in 1/m
_const_a0 = _const_alpha/(4*pi*_const_rydberg) # bohr radius in m
_const_proton_mass = gummy(1.672621898e-27,0.000000021e-27) # in kg

# constants from IAU 2009
_const_earth_mass = gummy(3.986004418e14,8e5)/_const_G # in kg
_const_solar_mass = gummy(1.32712442099e20,1e10)/_const_G # in kg
_const_jupiter_mass = _const_solar_mass/gummy(1.047348644e3,1.7e-5) # in kg

with _BuiltinLib():
    #SI Base units
    _m = PrefixedUnit('metre','m',additional_names=('meter',),add_symbol=True,
                      order=1,description='SI unit of length',
                      base_description='SI base unit of length')
    Unit.alias('micron','um')
    _cm = Unit.unit('cm')
    _cm.description = 'SI unit of length and CGS base unit'
    
    _kg = PrefixedUnit('gram','g',add_symbol=True,order=1,base_prefix='kilo',
                      description='SI unit of mass',
                      base_description='SI base unit of mass')
    Unit.unit('g').description = 'SI unit of mass and CGS base unit'
    
    _s = PrefixedUnit('second','s',add_symbol=True,order=3,
                      description='SI unit of time',
                      base_description='SI and CGS base unit of time')
    
    _A = PrefixedUnit('ampere','A',add_symbol=True,order=4,
                      description='SI unit of electrical current',
                      base_description='SI base unit of electrical current')
    
    _K = PrefixedUnit('kelvin','K',add_symbol=True,order=5,
                      description='SI unit of temperature',
                      base_description='SI base unit of temperature')
    
    _mol = PrefixedUnit('mole','mol',add_symbol=True,order=6,
                      description='SI unit for amount of substance',
                      base_description='SI base unit for amount of substance')
    
    _cd = PrefixedUnit('candela','cd',add_symbol=True,order=7,
                      description='SI unit of luminosity',
                      base_description='SI base unit of luminosity')
    
    _ipk = Unit('international prototype kilogram','m(K)',
                Conversion(_kg,gummy(1,1.2e-8)),add_symbol=True,
                html_symbol='<i>m</i>(&#x1d4a6;)',latex_symbol='m({\mathcal {K}})',
                description='IPK, le grand k')
    Unit.alias('IPK',_ipk)
    
    #Coherent derived units in the SI with special names and symbols
    _rad = PrefixedUnit('radian','rad',Conversion(_m*_m**-1,1),add_symbol=True,order=0,
                        description='unit of angle, angular unit')
    _sr = PrefixedUnit('steradian','sr',Conversion(_m**2*_m**-2,1),add_symbol=True,order=0,
                       description='unit of solid angle, angular unit')
    _Hz = PrefixedUnit('hertz','Hz',Conversion(_s**-1,1),add_symbol=True,order=0,
                       description='frequency')
    _N = PrefixedUnit('newton','N',Conversion(_m*_kg*_s**-2,1),add_symbol=True,order=0,
                      description='SI derived unit of force')
    _Pa = PrefixedUnit('pascal','Pa',Conversion(_N*_m**-2,1),add_symbol=True,order=0,
                       description='SI derived unit of pressure')
    _J = PrefixedUnit('joule','J',Conversion(_N*_m,1),add_symbol=True,order=0,
                      description='SI derived unit of energy')
    _W = PrefixedUnit('watt','W',Conversion(_J*_s**-1,1),add_symbol=True,order=0,
                      description='SI derived unit of power')
    _C = PrefixedUnit('coulomb','C',Conversion(_s*_A,1),add_symbol=True,order=0,
                      description='SI derived unit of electrical charge')
    _V = PrefixedUnit('volt','V',Conversion(_W*_A**-1,1),add_symbol=True,order=0,
                      description='SI derived unit of electric potential, voltage')
    PrefixedUnit('farad','F',Conversion(_C*_V**-1,1),add_symbol=True,order=0,
                 description='SI derived unit of electrical capacitance')
    _ohm = PrefixedUnit('ohm','\u03A9',Conversion(_V*_A**-1,1),add_symbol=True,order=0,
                 html_symbol='&Omega;',latex_symbol=r'\Omega',ascii_symbol='ohm',
                 description='SI derived unit of electrical resistance')
    _siemens = PrefixedUnit('siemens','S',Conversion(_A*_V**-1,1),add_symbol=True,order=0,
                 description='SI derived unit of electrical conductance, susceptance and admittance')
    Unit.alias('mho',_siemens)
    _Wb = PrefixedUnit('weber','Wb',Conversion(_V*_s,1),add_symbol=True,order=0,
                       description='SI derived unit of magnetic flux')
    _T = PrefixedUnit('tesla','T',Conversion(_Wb*_m**-2,1),add_symbol=True,order=0,
                      description='SI derived unit of magnetic flux density')
    PrefixedUnit('henry','H',Conversion(_Wb*_A**-1,1),add_symbol=True,order=0,
                 description='SI derived unit of electrical inductance')
    
    _degC = OffsetUnit('degree Celsius','\u00B0C',OffsetConversion(_K,273.15),
               latex_symbol='^{\circ}C',ascii_symbol = 'degC',add_symbol=True,order=0,
               description='unit of temperature')
    Unit.alias('degree C',_degC)
    Unit.alias('deg C',_degC)
    
    _lm = PrefixedUnit('lumen','lm',Conversion(_cd*_sr,1),add_symbol=True,order=0,
                       description='SI derived unit for luminous flux')
    PrefixedUnit('lux','lx',Conversion(_lm*_m**-2,1),add_symbol=True,order=0,
                 description='SI derived unit for illuminance and luminous emittance')
    PrefixedUnit('becquerel','Bq',Conversion(_s**-1,1),add_symbol=True,order=0,
                 description='SI derived unit for radioactivity')
    PrefixedUnit('gray','Gy',Conversion(_J*_kg**-1,1),add_symbol=True,order=0,
                 description='SI derived unit for ionizing radiation')
    PrefixedUnit('sievert','Sv',Conversion(_J*_kg**-1,1),add_symbol=True,order=0,
                 description='SI derived unit for ionizing radiation dose')
    PrefixedUnit('katal','kat',Conversion(_mol*_s**-1,1),add_symbol=True,order=0,
                 description='SI unit for catalytic activity')
    
    # 1990 conventional unit definitions
    _KJ90 = MFraction('483597.9e9') # Josephson constant Hz/V, 1990 conventional value
    _KJ = 2*_const_e/_const_h
    _RK90 = MFraction('25812.807') # Von Klitzing constant in ohm, 1990 conventional value
    _RK = _const_h/_const_e**2
    _V90 = PrefixedUnit('volt 90','V(90)',Conversion(_V,_KJ90/_KJ),add_symbol=True,
                        html_symbol='V<sub>90</sub>',latex_symbol='V_{90}',
                        description='1990 conventional definition of voltage, electrical potential')
    _ohm90 = PrefixedUnit('ohm 90','\u03A9(90)',Conversion(_ohm,_RK/_RK90),add_symbol=True,
                          html_symbol='&Omega;<sub>90</sub>',latex_symbol=r'\Omega_{90}',ascii_symbol='ohm(90)',
                          description='1990 conventional definition of electrical resistance')
    _A90 = PrefixedUnit('ampere 90','A(90)',Conversion(_V90*_ohm90**-1,1),add_symbol=True,
                        html_symbol='A<sub>90</sub>',latex_symbol='A_{90}',
                        description='1990 conventional definition of electrical current')
    PrefixedUnit('coulomb 90','C(90)',Conversion(_s*_V90*_ohm90**-1,1),add_symbol=True,
                 html_symbol='C<sub>90</sub>',latex_symbol='C_{90}',
                 description='1990 conventional definition of electrical charge')
    PrefixedUnit('watt 90','W(90)',Conversion(_V90*_A90,1),add_symbol=True,
                 html_symbol='W<sub>90</sub>',latex_symbol='W_{90}',
                 description='1990 conventional definition of electrical power')
    PrefixedUnit('farad 90','F(90)',Conversion(_s*_ohm90**-1,1),add_symbol=True,
                 html_symbol='F<sub>90</sub>',latex_symbol='F_{90}',
                 description='1990 conventional definition of electrical capacitance')
    PrefixedUnit('henry 90','H(90)',Conversion(_s*_ohm90,1),add_symbol=True,
                 html_symbol='H<sub>90</sub>',latex_symbol='H_{90}',
                 description='1990 conventional definition of electrical inductance')
    
    # 1954 Kelvin definition
    PrefixedUnit('kelvin 54','K(54)',Conversion(_K,gummy(1,u=3.7e-7)),add_symbol=True,
                 html_symbol='K<sub>54</sub>',latex_symbol='K_{54}',
                 description = 'unit of temperature, 1954 SI definition based on the triple point of water')
    
    #Non-SI units accepted for use with the International System of Units
    _min = Unit('minute','min',Conversion(_s,60),add_symbol=True,order=0,
                description='unit of time')
    _h = Unit('hour','h',Conversion(_min,60),add_symbol=True,order=0,
              description='unit of time')
    _d = Unit('day','d',Conversion(_h,24),add_symbol=True,order=0,
              description='unit of time')
    Unit.alias('D',_d)
    _deg = Unit('degree','\t\u00B0',Conversion(_rad,pi/180),add_symbol=True,order=0,
                 latex_symbol='^{\circ}',ascii_symbol='deg',
                 description='unit of angle, angular unit')
    _arcmin = Unit('arcminute',"\t'",Conversion(_deg,MFraction(1,60)),add_symbol=True,order=0,
                   short_name='arcmin',description='unit of angle, angular unit')
    Unit('arcsecond','\t"',Conversion(_arcmin,MFraction(1,60)),add_symbol=True,order=0,
         short_name='arcsec',description='unit of angle, angular unit')
    Unit('hectare','ha',Conversion('hm**2',1),add_symbol=True,order=0,
         description='unit of area')
    PrefixedUnit('litre','L',Conversion('dm**3',1),add_symbol=True,order=0,
                 additional_names=('liter',),description='unit of volume')
    _tonne = PrefixedUnit('tonne','t',Conversion(_kg,1000),add_symbol=True,order=0,
                          prefixes=['centi','deci','deca','hecto','kilo','mega',
                                    'giga','tera','peta','exa','zetta','yotta'],
                          description='unit of mass')
    Unit.alias('metric ton',_tonne)
    
    _Da = PrefixedUnit('dalton','Da',Conversion(_kg,_const_dalton),
                 add_symbol=True,description='unified atomic mass unit')
                 
    #Other non-SI units
    PrefixedUnit('bar','bar',Conversion(_Pa,100000),add_symbol=True,order=0,
                 description='unit of pressure')
    Unit('millimetre of mercury','mmHg',Conversion(_Pa,133.322387415),add_symbol=True,
         order=0,description='unit of pressure')
    Unit('angstrom','\u212B',Conversion(_m,MFraction('1e-10')),add_symbol=True,order=0,
         html_symbol='&#8491;',ascii_symbol='angstrom',description='unit of length')
    
    _nm = Unit('nautical mile','M',Conversion(_m,1852),add_symbol=True,order=0,
               description='unit of length')
    Unit.alias('NM',_nm)
    Unit.alias('nmi',_nm)
    Unit.alias('Nm',_nm)
    
    Unit('barn','b',Conversion(_m**2,MFraction('1e-28')),add_symbol=True,order=0,
         description='unit of area')
    Unit('knot','kn',Conversion(_nm*_h**-1,1),add_symbol=True,order=0,
         description='unit of speed')
    _dbp = LogUnit('decibel','dB',LogConversion(1,10,10,_lg10),add_symbol=True)
    Unit.alias('dB power',_dbp)
    Unit.alias('dB-p',_dbp)
    
    _dbf = LogUnit('decibel field','dB',LogConversion(1,20,10,_lg10),short_name='dB-f')
    Unit.alias('dB field',_dbf)
    Unit.alias('dB root-power',_dbf)
    
    def _ln(x):
        try:
            return np.log(x)
        except:
            return np.log(float(x))
    LogUnit('neper','Np',LogConversion(1,1,e,_ln),add_symbol=True)
    
    #Non-SI units associated with the CGS and the CGS-Gaussian system of units
    PrefixedUnit('erg','erg',Conversion(_J,MFraction('1e-7')),add_symbol=True,order=0,
                 description='CGS unit of energy')
    _dyn = PrefixedUnit('dyne','dyn',Conversion(_N,MFraction('1e-5')),add_symbol=True,order=0,
                        description='CGS unit of force')
    PrefixedUnit('poise','P',Conversion('dyn s cm**-2',1),add_symbol=True,order=0,
                 description='CGS unit of dynamic viscosity')
    PrefixedUnit('stokes','St',Conversion('cm**2 s**-1',1),add_symbol=True,order=0,
                 description='CGS unit of kinematic viscosity')
    PrefixedUnit('stilb','sb',Conversion('cd cm**-2',1),add_symbol=True,order=0,
                 description='CGS unit of luminance')
    PrefixedUnit('phot','ph',Conversion('cd sr cm**-2',1),add_symbol=True,order=0,
                 description='CGS unit of illuminance or luminous flux')
    PrefixedUnit('galileo','Gal',Conversion('cm s**-2',1),add_symbol=True,order=0,
                 description='CGS unit of acceleration')
    PrefixedUnit('maxwell','Mx',Conversion(_Wb,MFraction('1e-8')),add_symbol=True,order=0,
                 description='CGS unit of magnetic flux')
    PrefixedUnit('gauss','G',Conversion(_T,MFraction('1e-4')),add_symbol=True,order=0,
                 description='CGS unit of magnetic flux density')
    PrefixedUnit('oersted','Oe',Conversion(_A*_m**-1,1000/(4*pi)),add_symbol=True,
                 order=0,description='CGS unit of auxiliary magnetic field')
    
    # decibel units
    _dBSPL = LogUnit('decibel sound pressure level','dB',
                 LogConversion(gummy(20,unit='uPa'),20,10,_lg10),
                 short_name='dB(SPL)',add_symbol=False,
                 description='sound pressure level in air')
    Unit.alias('dB(20uPa)',_dBSPL)
    
    _dBuPa = LogUnit('decibel \u03BCPa','dB',
                   LogConversion(gummy(1,unit='uPa'),20,10,_lg10),
                   short_name='dB(uPa)',add_symbol=False,
                   description='sound pressure level in water')
    
    _dBSIL = LogUnit('decibel sound intensity level','dB',
                 LogConversion(gummy(1e-12,unit=_W/_m**2),10,10,_lg10),
                 short_name='dB(SIL)',add_symbol=False)
    
    _dBSWL = LogUnit('decibel sound power level','dB',
                 LogConversion(gummy(1e-12,unit=_W),10,10,_lg10),
                 short_name='dB(SWL)',add_symbol=False)
    Unit.alias('dB(pW)',_dBSWL)
    
    _dBm = LogUnit('decibel milliwatt','dBm',LogConversion(gummy(1,unit='mW'),10,10,_lg10),
                   add_symbol=True,description='unit of power')
    Unit.alias('dB(mW)',_dBm)
    
    _dBV = LogUnit('decibel volt','dBV',LogConversion(gummy(1,unit=_V),20,10,_lg10),
                   add_symbol=True,description='unit of voltage')
    Unit.alias('dB(V)',_dBV)
    
    _dBv = LogUnit('decibel u','dBu',LogConversion(gummy(sqrt(0.6),unit=_V),20,10,_lg10),
                   add_symbol=True,
                   description='unit of voltage, volt, decibel relative to sqrt(0.6) V')
    Unit.alias('dBu',_dBv)
    
    _dBmV = LogUnit('decibel millivolt','dBmV',LogConversion(gummy(1,unit='mV'),20,10,_lg10),
                    add_symbol=True,description='unit of voltage, volt')
    Unit.alias('dB(mV)',_dBmV)
    
    _dBuV = LogUnit('decibel microvolt','dB\u03bcV',LogConversion(gummy(1,unit='uV'),20,10,_lg10),
                    add_symbol=True,ascii_symbol='dBuV',description='unit of voltage')
    Unit.alias('dB(uV)',_dBuV)
    
    _dBZ = LogUnit('decibel Z','dBZ',LogConversion(gummy(1,unit='mm**6 m**-3'),10,10,_lg10),
                    add_symbol=True,description='unit used with weather radar')
    Unit.alias('dB(Z)',_dBZ)
    
    _dBJ = LogUnit('decibel joule','dBJ',LogConversion(gummy(1,unit=_J),10,10,_lg10),
                    add_symbol=True,description='unit of energy')
    Unit.alias('dB(J)',_dBJ)
    
    _dBmu = LogUnit('decibel microvolt per metre','dB\u03bc',LogConversion(gummy(1,unit='uV/m'),20,10,_lg10),
                    add_symbol=True,description='unit of voltage')
    Unit.alias('dB(uV/m)',_dBmu)
    Unit.alias('decibel microvolt per meter',_dBmu)
    
    _dBf = LogUnit('decibel femtowatt','dBf',LogConversion(gummy(1,unit='fW'),10,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('dB(fW)',_dBf)
    
    _dBW = LogUnit('decibel watt','dBW',LogConversion(gummy(1,unit=_W),10,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('dB(W)',_dBW)
    
    _dBk = LogUnit('decibel kilowatt','dBk',LogConversion(gummy(1,unit='kW'),10,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('dB(kW)',_dBk)
    
    _dBsm = LogUnit('decibel square metre','dBsm',LogConversion(gummy(1,unit=_m**2),10,10,_lg10),
                    add_symbol=True,description='unit used to measure antenna effective area')
    Unit.alias('dB(m**2)',_dBsm)
    Unit.alias('decibel square meter',_dBsm)
    
    _dBmm1 = LogUnit('decibel reciprocal metre','dB(m\u207b\u00b9)',LogConversion(gummy(1,unit=_m**-1),10,10,_lg10),
                    add_symbol=True,html_symbol='dB(m<sup>-1</sup>)',
                    latex_symbol='\t\t\\mathrm{dB}(\\mathrm{m}^{-1})',
                    ascii_symbol='dB(m**-1)',
                    description='unit used to measure antenna factor')
    Unit.alias('decibel reciprocal meter',_dBmm1)
    
    _dBHz = LogUnit('decibel hertz','dB-Hz',LogConversion(gummy(1,unit=_Hz),10,10,_lg10),
                    add_symbol=True,description='frequency')
    Unit.alias('dB(Hz)',_dBHz)
    
    _dBK = LogUnit('decibel kelvin','dBK',LogConversion(gummy(1,unit=_K),10,10,_lg10),
                    add_symbol=True,description='unit used to measure noise temperature')
    Unit.alias('dB(K)',_dBK)
    
    _dBmm1 = LogUnit('decibel reciprocal kelvin','dB(K\u207b\u00b9)',LogConversion(gummy(1,unit=_K**-1),20,10,_lg10),
                    add_symbol=True,html_symbol='dB(K<sup>-1</sup>)',
                    latex_symbol='\t\t\\mathrm{dB}(\\mathrm{K}^{-1})',
                    ascii_symbol='dB(K**-1)',description='temperature')
    
    _mBm = LogUnit('millibel milliwatt','mBm',LogConversion(gummy(1,unit='mW'),1000,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('mB(mW)',_mBm)
    
    LogUnit('bel','B',LogConversion(1,1,10,_lg10),add_symbol=False)
    
    _bit = BinaryPrefixedUnit('bit','bit',order=0,add_symbol=True,
                              additional_names=('shannon',),
                              description='unit of information')
    Unit.alias('shannon',_bit)
    
    _nat = Unit('natural unit of information','nat',Conversion(_bit,1/np.log(2)),
         add_symbol=True,order=0,description='unit of information')
    Unit.alias('nit',_nat)
    Unit.alias('nepit',_nat)
    
    _Hart = Unit('hartley','Hart',Conversion(_bit,np.log2(10)),add_symbol=True,
                 order=0,description='unit of information')
    Unit.alias('ban',_Hart)
    Unit.alias('dit',_Hart)
    
    BinaryPrefixedUnit('byte','B',Conversion(_bit,8),add_symbol=True,order=0,
                       description='unit of information')
    
    _nibble = Unit('nibble','nibble',Conversion(_bit,4),add_symbol=True,order=0,
                   description='unit of information')
    Unit.alias('nybble',_nibble)
    Unit.alias('nyble',_nibble)
    
    PrefixedUnit('torr','Torr',Conversion(_Pa,MFraction(101325,760)),add_symbol=True,
                         order=0,prefixes=['milli'],description='unit of pressure')
    Unit('standard atmosphere','atm',Conversion(_Pa,101325),add_symbol=True,order=0,description='unit of pressure')
    PrefixedUnit('electronvolt','eV',Conversion(_J,_const_e),add_symbol=True,order=0,description='unit of energy, the energy gained by one electron moving across one volt')
                 
    # astronomical units
    _c = Unit('speed of light','c',Conversion(_m/_s,_const_c),add_symbol=True,order=0,
         html_symbol='<i>c</i>',latex_symbol='\t\tc',
         description='natural unit of velocity')
    _au = Unit('astronomical unit','au',Conversion(_m,149597870700),add_symbol=True,
               order=0,description='astronomical unit of length')
    Unit.alias('ua',_au)
    PrefixedUnit('parsec','pc',Conversion(_au,648000/pi),add_symbol=True,order=0,
                 prefixes=['kilo','mega','giga'],
                 description='astronomical unit of length')
    _a = PrefixedUnit('Julian year','a',Conversion(_d,MFraction('365.25')),add_symbol=True,order=0,
                       description='astronomical unit of time',prefixes=['kilo','mega',
                       'giga','tera'],additional_names=('annum','year'),
                        additional_short_names=('yr',))
    Unit('light year','ly',Conversion(_c*_a,1),add_symbol=True,order=0,
         description='astronomical unit of length')
    Unit('siriometer','Sm',Conversion(_au,1000000),add_unit=False,
         description='astronomical unit of length')
    Unit('light second','light-second',Conversion(_c*_s,1),add_symbol=True,
         order=0,description='unit of length',)
    Unit('light minute','light-minute',Conversion(_c*_min,1),add_symbol=True,
         order=0,description='astronomical unit of length')
    Unit('light hour','light-hour',Conversion(_c*_h,1),add_symbol=True,
         order=0,description='astronomical unit of length')
    _M_solar = Unit('solar mass','M\u2609',Conversion(_kg,_const_solar_mass),
                    add_symbol=True,order=0,html_symbol='<i>M</i><sub>&#x2609;</sub>',
                    latex_symbol='\t\tM_{\odot}',ascii_symbol='M(solar)',
                    description='astronomical unit of mass, mass of sun')
    Unit.alias('M(Sun)',_M_solar)
    _M_J = Unit('Jupiter mass','M(J)',Conversion(_kg,_const_jupiter_mass),add_symbol=True,
         order=0,html_symbol='<i>M</i><sub>J</sub>',latex_symbol='\t\tM_{\mathrm{J}}',
         description='astronomical unit of mass') #check
    Unit.alias('M(Jup)',_M_J)
    Unit('Earth mass','M\u2295',Conversion(_kg,_const_earth_mass),add_symbol=True,
         order=0,html_symbol='<i>M</i><sub>&#x2295;</sub>',latex_symbol='\t\tM_{\oplus}',
         ascii_symbol='M(E)',description='astronomical unit of mass')
    _Jy = Unit('jansky','Jy',Conversion(_W*_m**-2*_Hz**-1,MFraction('1e-26')),add_symbol=True,
         decription='astronomical unit, spectral flux density, spectral irradiance')
    LogUnit('monochromatic AB magnitude','m(AB)',
            LogConversion(gummy(3631,unit=_Jy),-2.5,10,_lg10),
            add_symbol=True,html_symbol='<i>m</i><sub>AB</sub>',
            latex_symbol='\t\tm_{\\mathrm{AB}}',
            description='astronomical unit, spectral flux density, spectral irradiance')
    
    # natural units
    _hbar = Unit('natural unit of action','\u210f',Conversion(_J*_s,_const_hbar),
         order=0,add_symbol=True,ascii_symbol='h-bar')
    _e = Unit('elementary charge','e',Conversion(_C,_const_e),
         order=0,add_symbol=True,html_symbol='<i>e</i>',latex_symbol='\t\te',
         description='natural unit of charge, natural unit, atomic unit of charge')
    _a0 = Unit('bohr','a0',Conversion(_m,_const_a0),
         order=0,add_symbol=True,html_symbol='<i>a</i><sub>0</sub>',latex_symbol='\t\ta_{0}',
         description='natural unit of length, atomic unit of length, Bohr radius')
    Unit.alias('a(0)',_a0)
    
    _me = Unit('electon mass','m(e)',Conversion(_kg,_const_me),
         order=0,add_symbol=True,html_symbol='<i>m</i><sub>e</sub>',latex_symbol='\t\tm_{\\mathrm{e}}',
         description='natural, atomic unit of mass')
    _Eh = Unit('hartree','E_h',Conversion(_hbar**2*_me**-1*_a0**-2,1),order=0,
               add_symbol=True,html_symbol='<i>E</i><sub>h</sub>',latex_symbol='\t\tE_{\\mathrm{h}}',
               description='natural, atomic unit of energy')
    Unit.alias('Ha',_Eh)
    Unit('reduced compton wavelength','\u019b(C)',
         Conversion(_hbar*_me**-1*_c**-1,1),order=0,add_symbol=True,
         html_symbol='\u019b<sub>C</sub>',latex_symbol='\t\t\u019b_{\\mathrm{C}}',
         ascii_symbol='lambda(C)',
         description='electron reduced compton wavelength, natural unit of length')
    Unit('proton mass','m(p)',Conversion(_kg,_const_proton_mass),
         order=0,add_symbol=True,html_symbol='<i>m</i><sub>p</sub>',latex_symbol='\t\tm_{\\mathrm{p}}')

    Unit('Planck length','l(P)',Conversion(_m,sqrt(_const_hbar*_const_G*_const_c**-3)),order=0,
         add_symbol=True,html_symbol='<i>l</i><sub>P</sub>',
         latex_symbol='\t\tl_{\\mathrm{P}}',description='natural unit of length')
    Unit('Planck mass','m(P)',Conversion(_kg,sqrt(_const_hbar*_const_G**-1*_const_c)),order=0,
         add_symbol=True,html_symbol='<i>m</i><sub>P</sub>',
         latex_symbol='\t\tm_{\\mathrm{P}}',description='natural unit of mass')
    Unit('Planck time','t(P)',Conversion(_s,sqrt(_const_hbar*_const_G*_const_c**-5)),order=0,
         add_symbol=True,html_symbol='<i>t</i><sub>P</sub>',
         latex_symbol='\t\tt_{\\mathrm{P}}',description='natural unit of time')
    Unit('Planck charge','q(P)',Conversion(_C,_const_e/sqrt(_const_alpha)),order=0,
         add_symbol=True,html_symbol='<i>q</i><sub>P</sub>',
         latex_symbol='\t\tq_{\\mathrm{P}}',description='natural unit of charge')
    Unit('Planck temperature','T(P)',Conversion(_K,sqrt(_const_hbar*_const_G**-1*_const_c**5*_const_k**-2)),order=0,
         add_symbol=True,html_symbol='<i>T</i><sub>P</sub>',
         latex_symbol='\t\tT_{\\mathrm{P}}',description='natural unit of temperature')
    
    PrefixedUnit('volt ampere reactive','var',Conversion(_V*_A,1),add_symbol=True,
                 order=0,description='volt-ampere reactive power')
    PrefixedUnit('volt ampere','VA',Conversion(_V*_A,1),add_symbol=True,
                 order=0,description='volt-ampere apparent power')
    
    _M_W = LogUnit('moment magnitude','M(W)',
                   LogConversion(gummy(1,unit=_dyn*_cm),MFraction(2,3),10,_lg10,offset=-10.7),
                   add_symbol=True,html_symbol='<i>M</i><sub>W</sub>',
                   latex_symbol='\\textit{M}_{W}',
                   desription='moment magnitude scale, earthquake intensity')
    Unit.alias('MMS',_M_W)
    
    Unit.alias('wavenumber',_cm**-1)
    
    Unit('root hertz','\u221aHz',Conversion(_Hz**0.5,1),add_symbol=True,order=0,
         html_symbol='&radic;<span style="text-decoration:overline;">Hz</span>',
         latex_symbol='\t\t\\sqrt{\\mathrm{Hz}}',ascii_symbol='sqrtHz')
    



# -*- coding: utf-8 -*-

# module exceptions

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

class IncompatibleUnitsError(ValueError):
    """
    This exception is raised when an operation is attempted with gummys that
    have units that are incompatible for that operation.
    """
    pass

class UnitError(Exception):
    """
    Base class for Unit exceptions.
    """
    pass

class UnitLibError(UnitError):
    """
    This exception is raised when the UnitLibrary cannot parse a unit string.
    """
    pass

class UnitNotFoundError(UnitLibError):
    pass

class CircularUnitConversionError(UnitError):
    pass

class NoUnitConversionFoundError(UnitError):
    pass

class UnitLibNotFoundError(UnitLibError):
    pass

class NoSimulatedDataError(Exception):
    pass

class UnitWarning(Warning):
    pass

class GummyWarning(Warning):
    pass

class FitWarning(Warning):
    pass

class BudgetWarning(Warning):
    pass

class UncertiantyPrecisionWarning(Warning):
    pass


# -*- coding: utf-8 -*-

# module printing

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module contains code to support pretty printing of gummys using LaTex and 
HTML.
"""

try:
    from IPython.display import display,Math,HTML,Markdown
    ipython_installed = True
except ImportError:
    ipython_installed = False


def print_html(text):
    if ipython_installed:
        display(HTML(text))
    else:
        raise NotImplementedError('ipython must be installed to print in HTML format')
        
def print_latex(text):
    if ipython_installed:
        display(Math(text))
    else:
        raise NotImplementedError('ipython must be installed to print in HTML format')
        
def print_markdown(text):
    if ipython_installed:
        display(Markdown(text))
    else:
        raise NotImplementedError('ipython must be installed to print in Markdown format')
        

        
    
# Called from _repr_latex_ or Ipyhton.display.Math and gummy.latex_math 
# Not sure if this is necessary; may depend on the version of Ipython.
def _latex_math(text):
    return '$ ' + text + ' $'
 
def _latex_norm(text):
    return '\\text{' + text + '}'

# Same as above for text used in plot titles and labels
def _latex_norm_plot(txt):
    txt = txt.replace(' ',' \\ ')
    return '\\mathrm{' + txt + '}'


def _find_printer(value):
    if value is None:
        return 'any_but_latex'
    value = value.lower().strip()
    if value not in ['latex','html','ascii','unicode','any','any_but_latex']:
        raise ValueError('printer ' + str(value) + ' is not recognized')
    return value
    
class MetaPrettyPrinter(type):
    @property
    def printer(self):
        """
        Get or set the prefered display printer.  This is a string with one of 
        the following values:
        
        "any", "latex", "html", "unicode", "ascii", or "any_but_latex"
        
        "any" will usually pick html or latex output when running in an IPython 
        console or Jupyter notebook and unicode otherwise.
        
        "any_but_latex" will usually pick html when running in an IPython 
        console or Jupyter notebook and unicode otherwise.
        
        "latex" and "html" are only available when running under IPython.  If 
        these printers are not available the display will default to "unicode".
        """
        return PrettyPrinter._printer
    @printer.setter
    def printer(self,value):
        PrettyPrinter._printer = _find_printer(value)
        
class PrettyPrinter(metaclass=MetaPrettyPrinter):
    # Base class that can be inherited to provide HTML and LaTeX ouput to an
    # ipython console or Jupyter notebook.
    
    # The inheriting class must define a tostring(self,fmt='unicode',**kwds)
    # method that accepts fmt values in ['html',latex','unicode','ascii'].
    
    break_on_printing_error = False
    
    # Set latex_math to None in the inherited class if the output of the latex
    # and _repr_latex should not be in LaTeX math mode.
    latex_math = _latex_math
    
    latex_norm = _latex_norm
    latex_math_plot = _latex_math
    latex_norm_plot = _latex_norm_plot
    
    _printer = 'any_but_latex'
            
    @property
    def printer(self):
        """
        Get or set the preferred display printer.  This is a string with one of
        the following values:
        
        "any", "latex", "html", "unicode", "ascii", or "any_but_latex"
        
        "any" will usually pick html or latex output when running in an IPython 
        console or Jupyter notebook and unicode otherwise.
        
        "any_but_latex" will usually pick html when running in an IPython 
        console or Jupyter notebook and unicode otherwise.
        
        "latex" and "html" are only available when running under IPython.  If 
        these printers are not available the display will default to "unicode".
        """
        return self._printer
    @printer.setter
    def printer(self,value):
        self._printer = _find_printer(value)
        
    def _repr_html_(self):
        if self._printer not in ['any','html','any_but_latex']:
            return None

        txt = self.tostring(fmt='html')

        return txt
            
    def _repr_latex_(self):
        if self._printer not in ['any','latex']:
            return None
        
        txt = self.tostring(fmt='latex')

        if self.latex_math is None:   
            return txt
        return type(self).latex_math(txt)
    
    def __str__(self):
        return self.tostring(fmt='unicode')
            
    def __repr__(self):
        if self. _printer == 'ascii':
            return self.tostring(fmt='ascii')
        return self.tostring(fmt='unicode')
            
    def __format__(self,fmt):
        return self.tostring(fmt=fmt)
    
    def tohtml(self,**kwds):
        """
        Returns a string representing this object formatted for html; equivalent
        to PrettyPrinter.tostring(fmt='html',**kwds).  See the tostring method.
        """
        return self.tostring(fmt='html',**kwds)
    
    def tolatex(self,**kwds):
        """
        Returns a string representing this object formatted for LaTeX; equivalent
        to PrettyPrinter.tostring(fmt='latex',**kwds).  See the tostring method.
        """
        return self.tostring(fmt='latex',**kwds)
    
    def tounicode(self,**kwds):
        """
        Returns a string representing this object; equivalent to 
        PrettyPrinter.tostring(fmt='unicode',**kwds) and PrettyPrinter.__str__().  
        See the tostring method.
        """
        return self.tostring(fmt='unicode',**kwds)
    
    def toascii(self,**kwds):
        """
        Returns a string representing this object formatted using only ASCII
        characters; equivalent to PrettyPrinter.tostring(fmt='ascii',**kwds).  
        See the tostring method.
        """
        return self.tostring(fmt='ascii',**kwds)
    
    def latex(self,math=None,**kwds):
        """
        Prints a representation of the object using LaTeX formatting if this method
        is called from an IPython console or Juptyer notebook. See the tostring
        method.
        """
        if math is None:
            math = type(self).latex_math
        
        if math is None:
            print_latex(self.tolatex(**kwds))
        else:
            print_latex(math(self.tolatex(**kwds)))
        
    def html(self,**kwds):
        """
        Prints a representation of the object using HTML formatting if this method
        is called from an IPython console or Juptyer notebook. See the tostring 
        method.
        """
        print_html(self.tohtml(**kwds))
        
    def unicode(self,**kwds):
        """
        Prints a representation of the object. Equivalent to 
        print(cls.tostring(fmt='unicode')).  See the tostring method.
        """
        print(self.tounicode(**kwds))
        
    def ascii(self,**kwds):
        """
        Prints a representation of the object using only ASCII characters.
        Equivalent to print(cls.tostring(fmt='ascii')).  See the tostring
        method.
        """
        print(self.toascii(**kwds))
    
    
def set_printer(value): 
    """
    Sets the preferred default display printer.  This is a string with one of
    the following values:
    
    "any", "latex", "html", "unicode", or "ascii"
    
    "any" will usually pick html or latex output when running in an IPython 
    console or Jupyter notebook and unicode otherwise.
    
    "latex" and "html" are only available when running under IPython.  If 
    these printers are not available the display will default to "unicode".
    """
    if value is None:
        PrettyPrinter._printer = 'any_but_latex'
        return
    value = value.lower().strip()
    if value not in ['latex','html','ascii','unicode','any','any_but_latex']:
        raise ValueError('printer ' + str(value) + ' is not recognized')
    PrettyPrinter._printer = value



# -*- coding: utf-8 -*-

# module nonlinearunit

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module defines the Nonlinear Unit and Nonlinear Conversion abstract 
classes that are the base class for the LogUnit and OffsetUnit classes.
"""

from .unit import Unit, Conversion
from .exceptions import IncompatibleUnitsError

class NonlinearConversion(Conversion):
    """
    Base class for non-linear conversions.
    """
    # self._unit must be set in __init__ and must be linear

    # self._parent is the owning unit and must be set in the owning unit's __init__
    
    linear = False
        
    def chain(self,c):
        return _ChainedConversion(self,c)
    
    def rchain(self,c):
        return _ChainedConversion(c,self)
        
    def copy(self):
        raise NotImplementedError('this is an "abstract" method and must be overriden in a subclass')
        
    def pow(self,e):
        if e != 1:
            raise ValueError()
        return self
        
    
class _ChainedConversion(NonlinearConversion):
    linear = False
    
    def __init__(self,c1,c2):
        self.parent = c1.parent
        self._unit = c2.unit
        
        if isinstance(c1,_ChainedConversion):
            frmlst = c1._frmlst
            tolst = c1._tolst
        elif not c1.linear:
            frmlst = [c1._frm]
            tolst = [c1._to]
        else:
            frmlst = []
            tolst = []
        
        if isinstance(c2,_ChainedConversion):
            frmlst = c2._frmlst + frmlst
            tolst += c2._tolst
        elif not c2.linear:
            frmlst = [c1._frm] + frmlst
            tolst += [c1._to]
            
        self._frmlst = frmlst
        self._tolst = tolst
        
        factor = 1
        if c1.linear or isinstance(c1,_ChainedConversion):
            factor *= c1.factor
        if c2.linear or isinstance(c2,_ChainedConversion):
            factor *= c2.factor
        self.factor = factor
        
        self._c1 = c1
        self._c2 = c2
                    
    def to(self,g):
        g = super().to(g)
        return g*self.factor
        
    def frm(self,g):
        g = g/self.factor
        return super().frm(g)
        
    def _to(self,g):
        for f in self._tolst:
            g = f(g)
        return g
        
    def _frm(self,g):
        for f in self._frmlst:
            g = f(g)
        return g
        
    def chain(self,c):
        return _ChainedConversion(self,c)
    
    def copy(self):
        return _ChainedConversion(self._c1,self._c2)
    

class NonlinearUnit(Unit):
    """Base class of non-linear units.
    """
    linear = False
        
    #def _getme(self,unit_list,power):
        # add this method to return a different instance depending on whether
        # the unit is in a composite unit or not. e.g. an OffsetUnit where
        # 'degC/m' actually means 'degC-i/m'.
        # This us called only from Unit.unit when returning a composite unit
        # from a parsed string.  unit_list is the raw unit list returned from 
        # the parser an power is the exponent for this unit.
        
    #def _format_xu(g,fmt,style,norm,nsig,solidus=solidus,mulsep=mulsep):
        # add this method to override the gummy display formatting
        # g is a gummy and fmt is a gummy format
        # return a tuple: ('override','put the string to be displayed here')
        
    def zero(self):
        raise NotImplementedError('this is an "abstract" method and must be overriden in a subclass')
        
    def get_composite(self,ul):
        # If this method raises a NotImplementedError then the unit cannot be
        # combined with other units.  Override this method to return a subclass
        # of _CompositeUnit to allow combinations.
        raise NotImplementedError()
        
    def from_uunit(self,u,unit):
        raise IncompatibleUnitsError('uunit may not be used with unit ' + self.tostring())
        
    def to_uunit(self,u,unit):
        raise IncompatibleUnitsError('uunit may not be used with unit ' + self.tostring())
    
    # override any of the methods below to allow the operations
    def _add(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the + operation is not defined with unit ' + self.tostring())
        
    def _radd(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the + operation is not defined with unit ' + self.tostring())
        
    def _sub(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the - operation is not defined with unit ' + self.tostring())
        
    def _rsub(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the - operation is not defined with unit ' + self.tostring())
        
    def _mul(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the * operation is not defined with unit ' + self.tostring())
        
    def _rmul(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the * operation is not defined with unit ' + self.tostring())
        
    def _truediv(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the / operation is not defined with unit ' + self.tostring())
        
    def _rtruediv(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the / operation is not defined with unit ' + self.tostring())
        
    def _pow(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the ** operation is not defined with unit ' + self.tostring())
        
    def _rpow(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the ** operation is not defined with unit ' + self.tostring())
        
    def _mod(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the mod operation is not defined with unit ' + self.tostring())
        
    def _rmod(self,a,bunit,b,aconv):
        raise IncompatibleUnitsError('the mod operation is not defined with unit ' + self.tostring())
        
    def _neg(self,a):
        raise IncompatibleUnitsError('the - operation is not defined with unit ' + self.tostring())
    
    def _pos(self,a):
        raise IncompatibleUnitsError('the + operation is not defined with unit ' + self.tostring())
    
    def _abs(self,a):
        raise IncompatibleUnitsError('the abs operation is not defined with unit ' + self.tostring())
        
    def __pow__(self,a):
        raise IncompatibleUnitsError('the pow operation is not defined with unit ' + self.tostring())
        
    def __mul__(self,a):
        raise IncompatibleUnitsError('the * operation is not defined with unit ' + self.tostring())
    
    def __truediv__(self,a):
        raise IncompatibleUnitsError('the / operation is not defined with unit ' + self.tostring())
    
    def __rtruediv__(self,a):
        raise IncompatibleUnitsError('the reciprocal of unit ' + self.tostring() + ' is not allowed')
        
class ReciprocalConversion(NonlinearConversion):
    def __init__(self,conversion):
        self._conversion = conversion
        if hasattr(conversion,'factor'):
            self.factor = 1/conversion.factor
        self._unit = 1/conversion._unit
        self.parent = 1/conversion.parent
        
    def to(self,g):
        return self._conversion.frm(g)
        
    def frm(self,g):
        return self._conversion.to(g)
        
    def chain(self,c):
        return _ChainedConversion(self,c)
    
    def copy(self):
        return ReciprocalConversion(self._conversion)


# -*- coding: utf-8 -*-

# module version

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

__version__ = '0.5.2'


# -*- coding: utf-8 -*-

import metrolopy as uc
import numpy as np
import warnings
from fractions import Fraction

try:
    import mpmath as mp
except:
    mp = None

rand = np.random.RandomState()

def test_gummy_init(n=None,exception_on_warning=True,prnt=False,plot=False):
    print('uc.gummy.p_method may be modified\nset uc.gummy.p_method = \'loc\' to recover the default value')
    from scipy.stats import norm, t
    
    if n is None:
        if prnt or plot:
            n = 100
        else:
            n=1000000
    
    units=['m','lb','m**2 s**3/kg**4','degF','dB(SPL)','%','Np']
    uunits = {'%':100, 'ppm':1e6, 'ppb':1e9, 'ms/s':1000}
    
    with warnings.catch_warnings():
        if exception_on_warning:
            warnings.simplefilter('error')
        else:
            warnings.simplefilter('ignore')
    
        for i in range(n):
            
            if rand.randint(2):
                unit = units[rand.randint(len(units))]
            else:
                unit = uc.one
                
            if rand.randint(2):
                if unit == 'm':
                    if rand.randint(2):
                        uunit = 'mm'
                    elif rand.randint(2):
                        uunit = 'm'
                    else:
                        uunit = 'in'
                else:
                    if unit in ['degF','dB(SPL)','Np']:
                        uunit = None
                    else:
                        uunit = list(uunits.keys())[rand.randint(len(uunits))]
            else:
                uunit = None
            
            bayesian = bool(rand.randint(2))
            
            if rand.randint(2):
                dof = rand.randint(3,20)
            else:
                dof = float('inf')
                
            if rand.randint(2):
                if rand.randint(2):
                    k = 4*rand.rand() + 0.1
                    p = None
                else:
                    p = rand.rand()/2 + 0.49
                    k = 1
            else:
                k = 1
                p = None
            
            if rand.randint(2):
                pmethods = ['loc','level of confidence','cp','coverage probability','gauss','ccp','ccp','conservative coverage probability','chebyshev']
                uc.gummy.p_method = pmethods[rand.randint(len(pmethods))]
            else:
                uc.gummy.p_method = None
            
            if unit == 'dB(SPL)':
                if rand.randint(10) == 0:
                    x = rand.randint(-10,10)
                else:
                    x = 100.0*(2*rand.rand() - 1)
            elif unit == 'Np':
                if rand.randint(10) == 0:
                    x = rand.randint(-10,10)
                else:
                    x = 20.0*(2*rand.rand() - 1)
            else:
                if rand.randint(10) == 0:
                    x = rand.randint(-60000,60000)
                else:
                    x = (2*rand.rand() - 1)*10.0**rand.randint(-10,10)
            if rand.randint(20) == 0:
                x = 0
                    
            if uunit in ['%','ppm','ppb','ms/s'] and not (unit == '%' and uunit == '%'):
                U = 1.0e3*rand.rand()
                u = abs(x)*U/uunits[uunit]
            elif unit == 'm' and uunit == 'mm':
                if x == 0:
                    U = 1.0e3*rand.rand()
                else:
                    U = abs(x)*1.0e3*rand.rand()
                u = U/1000
            elif unit == 'm' and uunit == 'in':
                if x == 0:
                    U = 100*rand.rand()
                else:
                    U = abs(x)*100*rand.rand()
                u = U*0.0254 
            else:
                if x == 0:
                    U = 1.0e12*rand.rand()
                else:
                    if rand.randint(20) == 0:
                        U = 1e6*abs(x)*rand.rand()
                    elif rand.randint(20) == 0:
                        U = rand.randint(20)+1
                    else:
                        U = 1.1*abs(x)*rand.rand()
                u = U
                    
            if rand.randint(2):
                name = chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65)
            else:
                name = None
                
            if rand.randint(2):
                utype = chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65)
            else:
                utype = None
                
            if rand.randint(20) == 0:
                assert uc.gummy(x) == x
                continue
            
            if rand.randint(20) == 0:
                const = True
                U = 0
                u = 0
            else:
                if u == 0:
                    const = True
                else:
                    const = False
                
            uc.gummy.bayesian = bayesian
            if uunit is not None and rand.randint(2):
                uu = uc.gummy(U,unit=uunit)
                g = uc.gummy(x,u=uu,unit=unit,dof=dof,k=k,p=p,name=name,
                             utype=utype)
            else:
                g = uc.gummy(x,u=U,unit=unit,dof=dof,k=k,p=p,uunit=uunit,
                             name=name,utype=utype)
                
            if not const:
                u = u/g.k
                
            orig_g = g

            if rand.randint(10) == 0:
                n = rand.randint(12)
                if bayesian and n > 0:
                    dof = float('inf')
                if n == 0:
                    g = uc.gummy(g)
                    assert g.name is None
                elif n == 1:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        g = 1*g
                elif n == 2:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        g = g*1
                elif n == 3:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        g = g/1
                elif  n == 4:
                    if unit == 'degF':
                        v = uc.gummy(0,unit='degF-i')
                    elif unit is not uc.one or rand.randint(2):
                        v = uc.gummy(0,unit=unit)
                    else:
                        v = 0
                    g = g + v
                elif n == 5:
                    if unit == 'degF':
                        v = uc.gummy(0,unit='degF-i')
                    elif unit is not uc.one or rand.randint(2):
                        v = uc.gummy(0,unit=unit)
                    else:
                        v = 0
                    g = v + g
                elif n == 6:
                    if unit == 'degF':
                        v = uc.gummy(0,unit='degF-i')
                    elif unit is not uc.one or rand.randint(2):
                        v = uc.gummy(0,unit=unit)
                    else:
                        v = 0
                    g = g - v
                elif n == 7:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    elif unit in ['dB(SPL)','Np']:
                        g =  g + uc.gummy(0,unit=unit)
                    else:
                        g = g**1
                elif n == 8:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        g = +g
                elif n == 9:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        g = 2.1*g
                        g = g/2.1
                elif n == 10:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        if unit is not uc.one or rand.randint(2):
                            v = uc.gummy(3.5*x,u=0.45*u,unit=unit)
                        else:
                            v = 3.3*x
                        g = v - g
                        g = g - v
                        g = -g
                else:
                    if unit == 'degF':
                        g =  g + uc.gummy(0,unit='degF-i')
                    else:
                        if x >= 0:
                            g = abs(g)
                        else:
                            g = -abs(g)
    
                k = 1
                p = None
                reinit = True
                uunit = None
                name = None
            else:
                reinit = False
    
            if x != 0 and g.x != 0:
                assert abs(g.x - x)/abs(x) < 1e-10
            else:
                assert abs(g.x - x) < 1e-10
                
            if x != 0 and g._x != 0:
                assert abs(g._x - x)/abs(x) < 1e-10
            else:
                assert abs(g._x - x) < 1e-10
                            
            if unit is 'degF':
                assert g.unit in [uc.Unit.unit('degF'),uc.Unit.unit('degF-i')]
            else:
                assert g.unit is uc.Unit.unit(unit)
            assert g.name == name
            
            if const:
                assert g.const
                assert g._u == 0
                if unit is uc.one:
                    if x == 0:
                        assert abs(g - x) < 1e-10
                    else:
                        assert abs(g - x)/x < 1e-10
                elif unit == '%':
                    if x == 0:
                         assert abs(g - x/100) < 1e-10
                    else:
                        assert abs((g - x/100)/(x/100)) < 1e-10
                elif unit == 'Np':
                    assert abs(g.convert(1) - np.exp(x))/np.exp(x) < 1e-10
                continue
            
            if uunit in ['%','ppm','ppb','dB','ms/s'] and unit != uunit:
                assert g.uunit_is_rel
            else:
                assert not g.uunit_is_rel
                
            assert (g._x - x)/u < 1e-6
            
            assert not g.const
            
            if dof == float('inf'):
                assert g.dof == float('inf')
            else:
                assert (g.dof - dof)/dof < 1e-6
            
            if uunit is not None and uunit != unit:
                assert g.uunit is uc.Unit.unit(uunit)
                
            if p is None:
                assert g.k == k
            else:
                assert g.p == p
            
            if uc.gummy.p_method in ['loc','level of confidence',None]:
                if dof == float('inf'):
                    assert abs(g.k - norm.ppf(0.5 + g.p/2)) < 1e-6
                else:
                    if bayesian:
                        if g.p is not None:
                            assert abs(g.k - np.sqrt((dof-2)/dof)*t.ppf(0.5 + g.p/2,dof)) < 1e-6
                    else:
                        assert abs(g.k - t.ppf(0.5 + g.p/2,dof)) < 1e-6
                        
            elif uc.gummy.p_method in ['cp','coverage probability','gauss']:
                if dof == float('inf') or bayesian:
                    if g.p is not None:
                        assert abs(g.k - 2/(3*(np.sqrt(1 - g.p)))) < 1e-6
                else:
                    if g.p is not None:
                        assert abs(g.k - np.sqrt(dof/(dof-2))*(2/(3*(np.sqrt(1 - g.p))))) < 1e-6                 
            else:
                if dof == float('inf') or bayesian:
                    if g.p is not None:
                        assert abs(g.k - 1/np.sqrt(1 - g.p)) < 10.0**-6
                else:
                    if g.p is not None:
                        assert abs(g.k - np.sqrt(dof/(dof-2))*(1/np.sqrt(1 - g.p))) < 1e-6
                    
            assert abs((g._u - u)/u) < 1e-10
            
            if not reinit:
                assert abs(g.U - U)/U < 1e-10
            
            assert abs((g._u - u)/u) < 1e-10
            
            assert '?' not in g.tostring()
            
            if prnt:
                print('------')
                print()
                if rand.randint(2):
                    styles = ['pm','pmi','concise','ueq','x','xf','u','uf','xunit','uunit']
                    g.style = styles[rand.randint(10)]
                    
                    g.show_k = bool(rand.randint(2))
                    g.show_dof = bool(rand.randint(2))
                    g.show_p = bool(rand.randint(2))
                    g.show_name = bool(rand.randint(2))
                    g.solidus = bool(rand.randint(2))
                    g.mulsep = bool(rand.randint(2))
                    if not rand.randint(5):
                        g.nsig = rand.randint(5) + 1
                    
                    assert '?' not in g.tostring()
                    print(g.style,', show_k=',g.show_k,', show_dof=',g.show_dof,', show_p=',g.show_p,', show_name=',g.show_name,', solidus=',g.solidus,', mulsep=',g.mulsep,', nsig=',g.nsig,':')
                    print()
                display(g)
                print()
                    
            if plot:
                print('------')
                print()
                print(g)
                g.sim()
                if rand.randint(2):
                    styles = ['pmsim','pmsimi','mcisim','cisim','usim','ufsim']
                    g.style = styles[rand.randint(6)]
                g.slashaxis = bool(rand.randint(2))
                print(g.style,', slashaxis=',g.slashaxis,':')
                print()
                display(g)
                print()
                g.hist()
                print()
                
            uc.gummy.bayesian = False
            
def display(g):
    w = rand.randint(5)
    assert '?' not in g.tostring()
    if w == 0:
        print(g)
    elif w == 1:
        g.latex()
    elif w == 2:
        g.html()
    elif w == 3:
        g.unicode()
    else:
        g.ascii()
                
def make_gummy(sign=None,exp=None,uexp=-6,sometimes_small=True,dof=None,
               unit=None,units=None,mpf=False,allowzero=True,allowlargeu=True):
    
    utypes = ['A','B','D',None]
    if units is None:
        units = ['m','lb','m**2 s**3/kg**4','m**3','s**-2','1']
        
    if rand.randint(10) == 0:
        if exp is None:
            x = rand.randint(6000)
        else:
            x = rand.randint(int(10**exp))
    else:
        x = rand.rand()*0.9 + 0.1
        
        if exp is None:
            exp = rand.randint(1,6)
            if rand.randint(2):
                exp = -exp
                
        x = x*10**exp
    
    if sign is None:
        if rand.randint(2):
            x = -x
    else:
        x *= sign
        
    if not allowzero and x == 0:
        x = x + 1
        
    if allowlargeu and rand.randint(10) == 0:
        u = rand.randint(20) + 1
    else:
        u = rand.rand()*0.5 + 0.5
        if sometimes_small and not rand.randint(10):
            uexp -= 4
        u = abs(x)*u*10**uexp
    
    if mpf and mp is not None and not rand.randint(4):
        x = mp.mpf(x)
        if rand.randint(2):
            u = mp.mpf(u)
        
    if dof is None:
        if not rand.randint(4):
            dof = float('inf')
        else:
            dof = rand.randint(5,15)
            
    if unit is None:
        unit = units[rand.randint(len(units))]

    utype = utypes[rand.randint(len(utypes))]
    
    g = uc.gummy(x,u=u,dof=dof,unit=unit,utype=utype)
    
    return (g,x,u,dof,unit,utype)

def make_number(sign=None,gconst=True,exp=None,fionly=True,allowzero=True):
    q = rand.randint(4)
    if fionly and q > 1:
        q = 1
    if q == 0:
        if exp is not None:
            x = rand.randint(10**(exp+1))
        else:
            x = rand.randint(100000)
    if q == 1 or q == 2:
        x = rand.rand()*0.9 + 0.1
        if exp is None:
            exp = rand.randint(1,6)
            if rand.randint(2):
                exp = -exp

        x = x*10**exp
    if q == 2 and mp is not None:
        x = mp.mpf(x)
    elif q == 3:
        if exp is not None:
            a = rand.randint(10**(exp+1))
            b = rand.randint(10**(exp+1))
        else:
            a = rand.randint(100000)
            b = rand.randint(100000)
        if b == 0:
            x = Fraction(0)
        else:
            x = Fraction(a,b)
        
    
    if sign is None:
        if rand.randint(2):
            x = -x
    else:
        x *= sign
    
    if gconst and rand.randint(2):
        x = uc.gummy(x)
    
    if not allowzero and x == 0:
        x = x + 1
        
    return x

def binary_func(f,df,sim=False,exp=None,fionly=False,uexp=-6,allowazero=True,
                allowbzero=True,allowlargeu=True):
    if sim:
        dof = dof=float('inf')
    else:
        dof = None
    a,ax,au,adof,_,_ = make_gummy(unit=1,dof=dof,exp=exp,mpf=(not fionly),
                                  uexp=uexp,allowzero=allowazero,
                                  allowlargeu=allowlargeu)
    if rand.randint(2):
        b = make_number(exp=exp,fionly=fionly,allowzero=allowbzero)
        if isinstance(b,uc.gummy):
            bx = b.x
        else:
            bx = b
        bu = 0
        bdof = float('inf')
    else:
        b,bx,bu,bdof,_,_ = make_gummy(unit=1,dof=dof,exp=exp,mpf=(not fionly),
                                      uexp=uexp,allowzero=allowbzero,
                                      allowlargeu=allowlargeu)
        
    ax = float(ax)
    bx = float(bx)
    au = float(au)
    bu = float(bu)
    x = f(ax,bx)
    d = df(ax,bx)
    da = float(d[0])
    db = float(d[1])
    u = np.sqrt((da*au)**2 + (db*bu)**2)
    if dof is None:
        dof = ((da*au)**4/adof + (db*bu)**4/bdof)
        if dof == 0:
            dof = float('inf')
        else:
            dof = u**4/dof
            
        if dof > uc.gummy.max_dof:
            dof = float('inf')
        if dof < 1:
            dof = 1
            
        dof = float(dof)
            
    g = f(a,b)
    
    if x == 0:
        assert abs(g.x) < 1e-15
    else:
        assert abs((g.x - x)/x) < 1e-14
        
    if u == 0:
        assert abs(u) < 1e-6
    else:
        assert abs((g.u - u)/u) < 1e-3
    
    if dof is not None and not np.isinf(dof):
        assert abs((g.dof - dof)/dof) < 0.01
    
    if sim and x != 0 and abs(u/x) > 1e-10:
        uc.gummy.simulate([a,b,g])
        
        assert abs((g.xsim - x)/(g.u/np.sqrt(1e5))) < 5
        assert abs((g.u - g.usim)/g.u) < 0.1
        
        if g.correlation(a) > 0.1:
            assert abs((g.correlation_sim(a)-g.correlation(a))/g.correlation(a)) < 0.1
            
        if isinstance(b,uc.gummy) and g.correlation(b) > 0.1:
            assert abs((g.correlation_sim(b)-g.correlation(b))/g.correlation(b)) < 0.1
    
def test_add(n=1000,sim=False):
    def f(a,b):
        return a + b
    def df(ax,bx):
        return (1,1)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_sub(n=1000,sim=False):
    def f(a,b):
        return a - b
    def df(ax,bx):
        return (1,-1)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_mul(n=1000,sim=False):
    def f(a,b):
        return a*b
    def df(ax,bx):
        return (bx,ax)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_div(n=1000,sim=False):
    def f(a,b):
        return a/b
    def df(ax,bx):
        return (1/bx,-ax/bx**2)
    if sim:
        lu = False
    else:
        lu = True
    for i in range(n):
        binary_func(f,df,sim=sim,allowbzero=False,allowlargeu=lu)
        
def test_pow(n=1000,sim=False):
    def f(a,b):
        return abs(a)**b
    def df(ax,bx):
        ax = abs(ax)
        return (bx*ax**(bx-1),np.log(ax)*ax**bx)
    if sim:
        lu = False
    else:
        lu = True
    for i in range(n):
        binary_func(f,df,sim=sim,exp=0,allowazero=False,allowlargeu=lu)
        
def test_mod(n=1000,sim=False):
    def f(a,b):
        return a%b
    def df(ax,bx):
        return (1,np.sign(bx)*abs(ax//bx))
    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True)
    
def test_abs(n=1000,sim=False):
    def f(a,b):
        return abs(a)*abs(b)
    def df(ax,bx):
        return (abs(bx),abs(ax))
    if sim:
        lu = False
    else:
        lu = True
    for i in range(n):
        binary_func(f,df,sim=sim,allowlargeu=lu,allowazero=lu,
                    allowbzero=lu)
        
def test_neg(n=1000,sim=False):
    def f(a,b):
        return (-a)*b
    def df(ax,bx):
        return (-bx,-ax)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_pos(n=1000,sim=False):
    def f(a,b):
        return (+a)*b
    def df(ax,bx):
        return (-bx,-ax)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_sincos(n=1000,sim=False):
    def f(a,b):
        return uc.sin(a) + uc.cos(b)
    def df(ax,bx):
        return (uc.cos(ax),-uc.sin(bx))
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_npsincos(n=1000,sim=False):
    def f(a,b):
        return np.sin(a) + np.cos(b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True)
        
def test_ap1sincos(n=1000,sim=False):
    def f(a,b):
        return uc.gummy.apply(np.sin,np.cos,a) + uc.gummy.napply(np.cos,b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True,exp=0,allowlargeu=False)
        
def test_ap2sincos(n=1000,sim=False):
    def ff(a,b):
        return np.sin(a) + np.cos(b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    def f(a,b):
        return uc.gummy.apply(ff,df,a,b)

    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True)
        
def test_apnsincos(n=1000,sim=False):
    def ff(a,b):
        return np.sin(a) + np.cos(b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    def f(a,b):
        return uc.gummy.napply(ff,a,b)

    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True,exp=0,allowlargeu=False)
        
def test_addxmul(n=1000):
    def f(*x):
        r = x[0]
        for i in range(len(x)-1):
            r += x[i+1]
        return r
        
    for j in range(n):
        k = rand.randint(10) + 2
        x = make_gummy(unit=1)[0]
        a = f(*(k*[x]))
        b = k*x
        c = uc.gummy.apply(f,lambda *x: len(x)*[1],*(k*[x]))
        d = uc.gummy.napply(f,*(k*[x]))
        
        assert abs((a.x - b.x)/b.x) < 1e-14
        assert (a.u - b.u)/b.u < 1e-3
        assert a.correlation(b) == 1
        assert a.correlation(x) == 1
        
        assert abs((c.x - b.x)/b.x) < 1e-14
        assert (c.u - b.u)/b.u < 1e-3
        assert c.correlation(b) == 1
        
        assert abs((d.x - b.x)/b.x) < 1e-14
        assert (d.u - b.u)/b.u < 1e-3
        assert d.correlation(b) == 1
        
def test_mulxpow(n=1000):
    def f(*x):
        r = x[0]
        for i in range(len(x)-1):
            r *= x[i+1]
        return r
        
    for j in range(n):
        k = rand.randint(10) + 2
        x = make_gummy(unit=1)[0]
        a = f(*(k*[x]))
        b = x**k
        xx = x.x
        c = uc.gummy.apply(f,lambda *x: len(x)*[xx**(k-1)],*(k*[x]))
        d = uc.gummy.napply(f,*(k*[x]))
        
        assert abs((a.x - b.x)/b.x) < 1e-14
        assert abs((a.u - b.u)/b.u) < 1e-3
        assert a.correlation(b) == 1
        
        if k%2 == 0 and x.x < 0:
            assert a.correlation(x) == -1
        else:
            assert a.correlation(x) == 1
            
        assert abs((c.x - b.x)/b.x) < 1e-14
        assert abs((c.u - b.u)/b.u) < 1e-3
        assert c.correlation(b) == 1
        
        assert abs((d.x - b.x)/b.x) < 1e-14
        assert abs((d.u - b.u)/b.u) < 1e-3
        assert d.correlation(b) == 1
        
def test_mulxnpow(n=1000):
    def f(*x):
        r = 1/x[0]
        for i in range(len(x)-1):
            r *= 1/x[i+1]
        return r
        
    for j in range(n):
        k = rand.randint(10) + 2
        x = make_gummy(unit=1,allowzero=False,allowlargeu=False)[0]
        a = f(*(k*[x]))
        b = x**-k
        d = uc.gummy.napply(f,*(k*[x]))
        
        assert abs((a.x - b.x)/b.x) < 1e-14
        assert abs((a.u - b.u)/b.u) < 1e-3
        assert a.correlation(b) == 1
        
        if k%2 == 0 and x.x < 0:
            assert a.correlation(x) == 1
        else:
            assert a.correlation(x) == -1
        
        assert abs((d.x - b.x)/b.x) < 1e-14
        assert abs((d.u - b.u)/b.u) < 1e-3
        assert d.correlation(b) == 1

def test_ubreakdown(n=None,prnt=False,budget=False):
    
    if n is None:
        if prnt:
            n = 100
        else:
            n = 10000
        
    for i in range(n):
        k = rand.randint(1,6)
        
        if rand.randint(2):
            unt = '1'
        else:
            unt = None
        g,x,u,dof,unit,utype = zip(*[make_gummy(unit=unt) for y in range(k)])
        
        for i,y in enumerate(g):
            assert y.utype is utype[i]
        
        c = rand.rand()*0.9 + 0.1
        if rand.randint(2):
            c = -c
        
        def f(*p):
            r = c
            for y in p:
                r *= y
            return r
        
        def d(k,*p):
            r = c
            for i,y in enumerate(p):
                if i != k:
                    r *= y
            return r
        
        gr = f(*g)
        
        ut = None
        for q in g:
            if gr._ref is q._ref:
                ut = q.utype
        assert gr.utype is ut
        
        x = list(x)
        u = list(u)
        lb = 'lb' in str(gr.unit)
        kg = 'kg' in str(gr.unit)
        for i,un in enumerate(unit):
            if 'lb' in un and not lb:
                x[i] *= 0.45359237
                u[i] *= 0.45359237
            if 'kg' in un and not kg and lb:
                x[i] *= 0.45359237**4
                u[i] *= 0.45359237**4
                
        xr = f(*x)
        
        assert abs(gr.x - xr)/abs(xr) < 1e-10
        
        ur = 0
        for i,y in enumerate(u):
            ur += y**2*d(i,*x)**2
        ur = np.sqrt(ur)
        assert abs(gr.u - ur)/ur < 1e-10
        
        dofr = 0
        for i,(v,idof) in enumerate(zip(u,dof)):
            dofr += v**4*d(i,*x)**4/idof
        if dofr == 0:
            dofr = float('inf')
        else:
            dofr = ur**4/dofr
        if dofr > uc.gummy.max_dof:
            dofr = float('inf')
            
        # Due to the fact that any dof larger than gummy.mox_dof is rounded to 
        # infinity in intermediate calculations, gr.dof may not be exactly dofr
        # particularly for large dofr.  
        if dofr > 100:
            assert gr.dof > 100
        else:
            assert abs(gr.dof - dofr)/dofr < 1e-2
            
        if unt == '1':
            if k == 1:
                dr = lambda *x: d(0,*x)
            else:
                dr = lambda *x: [d(i,*x) for i in range(k)]
            ga = uc.gummy.apply(f,dr,*g)
            assert abs(ga.x - xr)/abs(xr) < 1e-10
            assert abs(ga.u - ur)/ur < 1e-6
            if dofr > 100:
                assert ga.dof > 100
            else:
                assert abs(ga.dof - dofr)/dofr < 1e-2
                
            gn = uc.gummy.napply(f,*g)
            assert abs(gn.x - xr)/abs(xr) < 1e-10
            assert abs(gn.u - ur)/ur < 1e3
            if dofr > 100:
                assert gn.dof > 100
            else:
                assert abs(gn.dof - dofr)/dofr < 1e-2
                
            if rand.randint(2):
                if rand.randint(2):
                    gf = ga
                else:
                    gf = gn
            else:
                gf = gr
        else:
            gf = gr
            
        gl = list(g)
        if not rand.randint(4):
            gl.append(make_gummy(unit=unt)[0])
            
        if rand.randint(2):
            e = gl[rand.randint(len(gl))]
        else:
            e = [gl[rand.randint(len(gl))] for i in range(rand.randint(len(gl)))]
            
        ua = 0
        ub = 0
        ud = 0
        ue = 0
        dofa = 0
        dofb = 0
        dofd = 0
        dofe = 0
        for i,(ig,iu,idof) in enumerate(zip(g,u,dof)):
            if gf.ufrom(ig) == 0:
                assert iu*d(i,*x)/ur < 1e-3
            else:
                if ig.utype == 'A':
                    ua += iu**2*d(i,*x)**2
                    dofa += iu**4*d(i,*x)**4/idof
                elif ig.utype == 'B':
                    ub += iu**2*d(i,*x)**2
                    dofb += iu**4*d(i,*x)**4/idof
                elif ig.utype == 'D':
                    ud += iu**2*d(i,*x)**2
                    dofd += iu**4*d(i,*x)**4/idof
                
                try:
                    if e is ig or ig in e:
                        ue += iu**2*d(i,*x)**2
                        dofe += iu**4*d(i,*x)**4/idof
                except TypeError:
                    pass
            
        ua = np.sqrt(ua)
        ub = np.sqrt(ub)
        ud = np.sqrt(ud)
        ue = np.sqrt(ue)
        
        if dofa == 0:
            dofa = float('inf')
        else:
            dofa  = ua**4/dofa
        if dofb == 0:
            dofb = float('inf')
        else:
            dofb  = ub**4/dofb
        if dofd == 0:
            dofd = float('inf')
        else:
            dofd  = ud**4/dofd
        if dofe == 0:
            dofe = float('inf')
        else:
            dofe  = ue**4/dofe
                    
        if ua == 0:
            assert gf.ufrom('A') == 0
        else:
            assert abs(gf.ufrom('A') - ua)/ur < 1e-3
        if ua/ur > 0.01:
            if dofa > 100 or gf.ufrom('A') == 0:
                assert gf.doffrom('A') > 100
            else:
                assert abs(gf.doffrom('A') - dofa)/dofa < 1e-2
            
        if ub == 0:
            assert gf.ufrom('B') == 0
        else:
            assert abs(gf.ufrom('B') - ub)/ur < 1e-3
        if ub/ur > 0.01:
            if dofb > 100 or gf.ufrom('B') == 0:
                assert gf.doffrom('B') > 100
            else:
                assert abs(gf.doffrom('B') - dofb)/dofb < 1e-2
            
        if ud == 0:
            assert gf.ufrom('D') == 0
        else:
            assert abs(gf.ufrom('D') - ud)/ur < 1e-3
        if ud/ur > 0.01:
            if dofd > 100 or gf.ufrom('D') == 0:
                assert gf.doffrom('D') > 100
            else:
                assert abs(gf.doffrom('D') - dofd)/dofd < 1e-2
            
        if ue == 0:
            assert gf.ufrom(e) == 0
        else:
            assert abs(gf.ufrom(e) - ue)/ur < 1e-3
        #if ue/ur > 0.01:
            #if dofe > 100 or gf.ufrom(e) == 0:
                #assert gf.doffrom(e) > 100
            #else:
                #assert abs(gf.doffrom(e) - dofe)/dofe < 1e-2
                
        if not rand.randint(4):
            gf.uunit = '%'
                
        styles = ['pm','pmi','concise','ueq','u','uf']
        gf.style = styles[rand.randint(6)]
                
        assert '?' not in gf.tostring()
        
        ub = {i.utype for i in g if i.utype is not None}
        if not rand.randint(8) and len(ub) > 1:
            ub.pop()
        ub = list(ub)
        
        gfub = gf.copy(formatting=True)
        if len(ub) > 0:
            gfub.ubreakdown = ub
        
        assert '?' not in gfub.tostring()
                
        if prnt:
            w = rand.randint(5)
            if w == 0:
                print(gf)
                print(gfub)
            elif w == 1:
                gf.latex()
                gfub.latex()
            elif w == 2:
                gf.html()
                gfub.html()
            elif w == 3:
                gf.unicode()
                gfub.unicode()
            else:
                gf.ascii()
                gfub.ascii()
        
        if budget:
            show_expanded_u = bool(rand.randint(2))
            show_subtotals = bool(rand.randint(2))
            
            k = None
            p = None
            if bool(rand.randint(2)):
                if bool(rand.randint(2)):
                    if bool(rand.randint(2)):
                        gf.k = 2
                    else:
                        k = 2
                else:
                    if bool(rand.randint(2)):
                        gf.p = 0.95
                    else:
                        p = 0.95
                        
            if bool(rand.randint(2)):
                gl[0].name = 'gl0'
                
            if bool(rand.randint(2)) and len(gl) > 1:
                gl[1].name = 'gl1'
                
            if bool(rand.randint(2)):
                gf.name = 'gf'
            
            if not bool(rand.randint(4)):
                description = ['description ' + str(i) for i in range(len(gl)+1)]
            else:
                description = None
                
            if not bool(rand.randint(4)):
                custom = [str(i) for i in range(len(gl)+1)]
                custom_heading = 'n'
            else:
                custom = None
                custom_heading = None
                
            show_s = None
            show_c = None
            show_d = None
            if bool(rand.randint(2)):
                show_s = bool(rand.randint(2))
                show_c = bool(rand.randint(2))
                show_d = bool(rand.randint(2))
                
            b = gf.budget(gl,show_subtotals=show_subtotals,
                          show_expanded_u=show_expanded_u,k=k,p=p,
                          description=description,show_s=show_s,show_c=show_c,
                          show_d=show_d,custom=custom,
                          custom_heading=custom_heading)
            
            w = rand.randint(5)
            if w == 0:
                print(b)
            elif w == 1:
                b.latex()
            elif w == 2:
                b.html()
            else:
                b.unicode()
        
        if prnt or budget:
            print()
            print('---')
            print()


# -*- coding: utf-8 -*-

# module distributions

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the classed that represent distributions for use with 
gummy objects
"""

from .exceptions import NoSimulatedDataError
#from .unit import Unit
import numpy as np

class Distribution:
    """
    Abstract base class for distributions used for Monte-Carlo uncertainty
    propagation.
    
    In a derived class define the following methods:
        
    random(n=None):  Return a numpy array of n values drawn from the distribution.
        If `n` is `None` then a single scalar value should be returned.  Preferably
        use, as a random number generator, the `numpy.RandomState` object accessed
        with the `Distribution.random_state` static method.
        
    x():  A scalar "center" of the distribution.  This is used to get the `x` value
        of a gummy defined with the distribution.
        
    u():  A scalar "standard uncertainty" of the distribution (usually the standard
        deviation).  This is used to get the u value of a gummy defined with
        the distribution.

    Attributes
    ----------
    simdata: numpy.ndarray or `None`
        Simulated data following a call from the `simulate` static method.

    isindependent: `bool`
        `False` if the class is a `Convolution`, `True` otherwise
        
    Example
    -------
    >>> class ChiSquaredDist(Distribution):
    ...    def __init__(self,dof):
    ...        self.dof = dof
    ...
    ...    def random(self,n=None):
    ...        return Distribution.random_state().chisquare(self.dof,n)
    ...
    ...    def x(self):
    ...        return self.dof
    ...
    ...    def u(self):
    ...        return 2*self.dof
    
    """
    _random_state = None
    simdata = None
    
    isindependent = True
    
    _called = []
    _mean = None
    _stdev =  None
    _simsorted = None
    _ci = None
    _cisym = None
    _cov = None
    _used = False
    
    @staticmethod
    def random_state():
        """
        Returns the `numpy.random.RandomState` object shared by all distributions.
        """
        if Distribution._random_state is None:
            Distribution._random_state = np.random.RandomState()
        return Distribution._random_state
    
    @staticmethod
    def set_seed(seed):
        """
        Sets the seed of the `numpy.random.RandomState` object shared by all
        distributions.
        """
        Distribution.random_state().seed(seed)
    
    @staticmethod
    def apply(f,*d):
        """
        Applies a function `f` to distribution(s) `d1`, `d2`, ..., `f`(`d1`,`d2`,...)
        and returns a distribution resulting from the convolution.

        Parameters
        ----------
        f: function
            Any function that takes an appropriate number of scalar values
            and returns a scalar.
        
        d1, d2, ...: `Distribution` or scalar
            the distributions to be used as arguments to f.  Fixed values
            (e.g. `float`) can also be used for some or all arguments.

        Returns
        -------
        `Convolution` or `float`:
            if any of `d1`, `d2`, ... is a `Distribution`, f(d1,d2,...)
            otherwise
        """
        if not any([isinstance(v,Distribution) for v in d]):
            return f(*d)
        return Convolution(f,*d)
        
    @staticmethod
    def simulate(distributions,n=100000,ufrom=None):
        """
        Generates simulated data for the desired distributions.

        Parameters
        ----------
        distributions: `list` of `Distribution`
            Each Distribution in the list will have the simulated data
            available in the `simdata` attribute. Data will be cleared
            from all Distribution instances not in the list.
        
        n:  int > 0, optional
            The number of samples to generate for each distribution.  The
            default is 100000
            
        ufrom:  `list` of `Distributions` or `None`
            If this parameter is not `None`, then only variables represented
            by distributions in this list will be allowed to vary.  All other
            variables will be held fixed at the distribution x value.
        """
        for d in Distribution._called:
            d.clear()
        Distribution._called = []

        for d in distributions:
            if isinstance(d,Distribution):
                d._simulate(int(n),ufrom)
            
        for d in Distribution._called:
            if d not in distributions:
                d.clear()
                Distribution._called.remove(d)
         
    @property
    def mean(self):
        """
        Returns the mean of the simulated data.

        Raises
        ------
        A `NoSimulatedDataError` will be raised if no simulated data is available
        from a call to `Distribution.simulate()`.
        """
        if self._mean is not None:
            return self._mean
        if self.simdata is None:
            raise NoSimulatedDataError('simulate must be called to generate some data')
        self._mean = np.mean(self.simdata)
        return self._mean
        
    @property
    def stdev(self):
        """
        Returns the standard deviation of the simulated data.

        Raises
        ------
        A NoSimulatedDataError will be raised if no simulated data is available 
        from a call to Distribution.simulate().
        """
        if self._stdev is not None:
            return self._stdev
        if self.simdata is None:
            raise NoSimulatedDataError('simulate must be called to generate some data')
        self._stdev = np.std(self.simdata,ddof=1)
        
        if np.isnan(self._stdev):
            if np.any(np.isinf(self.simdata)):
                self._stdev = float('inf')
                
        return self._stdev
        
    @property
    def simsorted(self):
        """
        numpy.ndarray, read-only
        
        Returns a sorted numpy array containing the simulated data values.

        Raises
        ------
        A NoSimulatedDataError will be raised if no simulated data is available 
        from a call to Distribution.simulate().
        """
        if self._simsorted is not None:
            return self._simsorted
        if self.simdata is None:
            raise NoSimulatedDataError('simulate must be called to generate some data')
        self._simsorted = np.sort(self.simdata)
        return self._simsorted
        
    def ci(self,p):
        """
        Returns the shortest interval that contains the fraction p of the simulated
        data values.

        Returns
        -------
        `tuple` of `float`:
            a tuple containing the lower and upper limits of the interval


        Raises
        ------
        `NoSimulatedDataError`:
            if no simulated data is available from a call to
            `Distribution.simulate`.

        See Also
        --------
        cisym
        """
        if p > 1:
            raise ValueError('p > 1')
        if p < 0:
            raise ValueError('p < 0')
        if self._ci is None:
            self._ci = {}
        elif p in self._ci:
            return self._ci[p]
        if self.simdata is None:
            raise NoSimulatedDataError('simulate must be called to generate some data')
        s = self.simsorted
        n = int((1-p)*len(self.simdata)+0.5)
        if n == 0:
            mn = np.argmin(s[-n:] - s[:n])
            lower = upper = (s[mn] + s[mn-1])/2
        else:
            mn = np.argmin(s[-n:] - s[:n])
            lower = s[mn]
            upper = s[mn-n]
        self._ci[p] = (lower, upper)
        return (lower, upper)
        
    def cisym(self,p):
        """
        Returns the interval that contains the fraction p of the simulated
        data values and with an equal number of values below the lower limit of
        the interval and above the upper limit of the interval.

        Returns
        -------
        `tuple` of `float`:
            a tuple containing the lower and upper limits of the interval

        Raises
        ------
        `NoSimulatedDataError`:
            if no simulated data is available from a call to
            `Distribution.simulate`.

        See Also
        --------
        ci
        """
        if p > 1:
            raise ValueError('p > 1')
        if p < 0:
            raise ValueError('p < 0')
        if self._cisym is None:
            self._cisym = {}
        elif p in self._cisym:
            return self._cisym[p]
        if self.simdata is None:
            raise NoSimulatedDataError('simulate must be called to generate some data')
        s = self.simsorted
        lower = np.percentile(s,50*(1-p))
        upper = np.percentile(s,50*(1+p))
        self._cisym = (lower, upper)
        return (lower, upper)
        
    def hist(self,hold=False,xlabel=None,ylabel='$ \\mathrm{probability\\:density} $',
             title=None,**kwds):
        """
        Generates a histogram from the simulated data.
        
        Parameters
        ----------
        hold:  `bool`, optional
            If this is `False`, pyplot.show() will be called before the method
            exits.  The default is `False`
            
        xlabel:  `str` or `None`, optional
            a label for the histogram horizontal axis, the default is `None`
        
        ylabel: `str` or `None`, optional
            a label for the histogram vertical axis, the default is 'probability
            density'
            
        title: `str` or `None`
            a title for the histogram, the default is `None`
        
        kwds:
            additional key words that will be passed to the `pyplot.hist` method
            that actually created the histogram

        Raises
        ------
        `NoSimulatedDataError`:
            if no simulated data is available from a call to
            `Distribution.simulate`.
        """
        import matplotlib.pyplot as plt
        if self.simdata is None:
            raise NoSimulatedDataError('simulate must be called to generate some data')
        if not 'bins' in kwds:
            kwds['bins'] = 100
        if not ('normed' in kwds or 'density' in kwds):
            try:
                from matplotlib import __version__ as version
                version = [int(v) for v in version.split('.')[:2]]
                usedensity = (version[0] > 2 or 
                              (version[0] == 2 and version[1] > 0))
            except:
                usedensity = True
            if usedensity:
                kwds['density'] = True
            else:
                kwds['normed'] = True
        if not 'histtype' in kwds:
            kwds['histtype'] ='stepfilled'
        plt.hist(self.simsorted[np.abs(self.simsorted) != np.inf],**kwds)
        if ylabel is not None:
            plt.ylabel(ylabel)
        if xlabel is not None:
            plt.xlabel(xlabel)
        if title is not None:
            plt.title(title)
        if not hold:
            plt.show()
                
    def clear(self):
        """
        Clears the simulated data.
        """
        self.simdata = None
        self._mean = None
        self._stdev = None
        self._simsorted = None
        self._ci = {}
        self._cisym = {}
        self._cov = {}
        
    def covsim(self,d):
        """
        Returns the covariance between this `Distribution` instance and another
        `Distribution` `d`.

        Raises
        ------
        `NoSimulatedDataError`:
            if no simulated data is available from a call to
            `Distribution.simulate` for both `self` and `d`.
        """
        if self._cov is None:
            self._cov = {}
        if d in self._cov:
            return self._cov[d]
        if self.simdata is None or d.simdata is None:
            raise NoSimulatedDataError('simulated data does not exist for both distributions')
        cov = np.cov([self.simdata,d.simdata])[1][0]
        self._cov[d] = cov
        return cov
        
    @staticmethod
    def covsim_matrix(*d):
        """
        Returns the variance-covariance matrix of the Distributions `d1`, `d2`, ...

        Raises
        ------
        `NoSimulatedDataError`:
            if no simulated data is available from a call to
            `Distribution.simulate` for any of `d1`, `d2`, ...
        """
        if any([(v.simdata is None) for v in d]):
            raise NoSimulatedDataError('simulated data does not exist for all distributions')
        return np.cov([v.simdata for v in d])
        
    @staticmethod
    def covplot(x,y,fmt='ko',xlabel=None,ylabel=None,title=None,hold=False,**kwds):
        """
        Plots the Distribution `x` versus the Distribution `y`

        Parameters
        ----------
        x, y:  `Distribution`
            the distributions to be plotted
            
        fmt:   `str`, optional
            Format parameter passed to `pyplot.plot()`, the default is 'ko'
        
        xlabel:  `str` or `None`
         a label for the plot x-axis, the default is `None`
        
        ylabel:  `str` or `None`
         a label for the plot y-axis, the default is `None`
            
        title: `str` or `None`
            a title for the histogram, the default is `None`
        
        hold:  `bool` i
            If this is `False`, `pyplot.show()` will be called before the
            method exits.  The default is `False`
            
        kwds:
            additional key words that will be passed to `pyplot.plot()`

        Raises
        ------
        `NoSimulatedDataError`:
            if no simulated data is available from a call to
            `Distribution.simulate` for `x` and `y`.
        """
        import matplotlib.pyplot as plt
        
        if x.simdata is None or y.simdata is None:
            raise NoSimulatedDataError('simulated data does not exist for both distributions')
        if 'ms' not in kwds and 'markersize' not in kwds:
            n = len(x.simdata)
            if n >= 100000:
                kwds['ms'] = 0.2
            elif n >= 50000:
                kwds['ms'] = 0.2
            elif n >= 20000:
                kwds['ms'] = 0.5
            else:
                kwds['ms'] = 1
        plt.plot(x.simdata,y.simdata,fmt,**kwds)
        if title is None:
            cov = Distribution.covsim_matrix(x,y)
            cor = cov[1][0]/np.sqrt(cov[1][1]*cov[0][0])
            if abs(cor) < 0.0005:
                cor = '0.000'
            elif abs(cor) < 0.1:
                cor = '{:.3f}'.format(cor)
            else:
                cor = '{:.2f}'.format(cor)
            title = '$ \\mathrm{correlation} = ' + cor + ' $'
            plt.title(title)
        if xlabel is None:
            xlabel = '$ x $'
        plt.xlabel(xlabel)
        if ylabel is None:
            ylabel = '$ y $'
        plt.ylabel(ylabel)
        if not hold:
            plt.show()
        
    def _simulate(self,n,ufrom):
        if ufrom is not None and self not in ufrom:
            self.simdata = np.full(n,self.x())
            return
        Distribution._called.append(self)
        self.simdata = self.random(n)
        
    def __add__(self,v):
        return Convolution(np.add,self,v)
        
    def __radd__(self,v):
        return Convolution(np.add,v,self)
        
    def __sub__(self,v):
        return Convolution(np.subtract,self,v)
        
    def __rsub__(self,v):
        return Convolution(np.subtract,v,self)
        
    def __mul__(self,v):
        return Convolution(np.multiply,self,v)
        
    def __rmul__(self,v):
        return Convolution(np.multiply,v,self)
        
    def __truediv__(self,v):
        return Convolution(np.divide,self,v)
        
    def __rtruediv__(self,v):
        return Convolution(np.divide,v,self)
        
    def __pow__(self,v):
        return Convolution(np.power,self,v)
        
    def __rpow__(self,v):
        return Convolution(np.power,v,self)
        
    def __abs__(self):
        return Convolution(np.abs,self)
        
    def __neg__(self):
        return Convolution(np.negative,self)
        
    def __pos__(self):
        return self
        
    def random(self,n=None):
        """
        Override this method in a derived class

        Return a numpy array of n values drawn from the distribution.
        If n is None then a single scalar value should be returned.
        """

        raise NotImplementedError()
        
    def x(self):
        """
        Override this method in a derived class

        Return a scalar "center" of the distribution.  This is used to
        get the `x` value of a gummy defined with the distribution.
        """
        raise NotImplementedError()
        
    def u(self):
        """
        Override this method in a derived class

        Return a scalar "standard uncertainty" of the distribution (usually
         the standard deviation).  This is used to get the `u` value of a
         gummy defined with the distribution.
        """
        # Override in a derived class.
        # A
        raise NotImplementedError()
    

class Convolution(Distribution):
    isindependent = False
    
    def __init__(self,func,*args):
        """
        Represents the distribution resulting from applying func to d1, d2, ...
        func(d1,d2,...)

        Parameters
        ----------
        func:  function
            a function that takes an appropriate number of scalar arguments and
            returns a scalar.
        
        *args: `Distribution` or `float`
            arguments for `func`
        """
        self.func = func
        self.args = args
        
    def random(self,n=None):
        raise NotImplementedError('use the simulate static method to generate data from a Convolution')
        
    def x(self):
        args = [a.x() if isinstance(a,Distribution) else a for a in self.args]
        return self.func(*args)
        
    def _random(self,n,ufrom):
        for a in self.args:
            if isinstance(a,Distribution) and (a not in Distribution._called):
                a._simulate(n,ufrom)
        v = [a.simdata if isinstance(a,Distribution) else a for a in self.args]
        
        if not isinstance(self.func,np.ufunc):
            # If func is not a numpy ufunc, see if it broadcasts like one
            try:
                ntst = max(n,self.args+5)
                vtst = [a.simdata[:ntst] if isinstance(a,Distribution) else a for a in self.args]
                tst = self.func(*vtst)
                if len(tst) != ntst:
                    raise TypeError()
                    
                func = self.func
            except:
                # if it doesn't broadcast, make it broadcast with frompyfunc
                func = np.frompyfunc(self.func,len(v),1)
        else:
            func = self.func
            
        ret = func(*v)

        if ret.dtype != np.float64:
            ret = np.array(ret,dtype=np.float64)
            
        return ret
                
    def _simulate(self,n,ufrom):
        Distribution._called.append(self)
        self.simdata = self._random(n,ufrom)
        

class MultivariateDistribution:
    def __init__(self,nd):
        """
        Base class for multivarate distributions.  `nd` is the number of dimensions
        of the distribution.
        
        To create a multi-variate distribution, inherit from the 
        MultivariateDistribution and define the following methods:

        _simulate(n):  Return a numpy array of n samples drawn from the distribution. 
            Preferably use, as a random number generator, the numpy RandomState 
            object accessed with the Distribution.random_state() static method.
                
        x():  a list or array with the "center" of the distribution for each 
            dimension.  This is used to get the x value of a gummys defined with 
            the distribution.
                
        u():  A list or array with "standard uncertainty" of the distribution 
            (usually the standard deviation) for each dimension.  This is used 
            to get the u value of a gummy defined with the distribution.
        
        and the following read-only property
        
        cov:  Returns the variance-covariance matrix
        
        The __init__ function must also call the MultivariateDistribution 
        __init__ with the number of dimensions nd of the distribution 
        e.g. super().__init__(nd).
        
        Example
        -------

                class DirechletDist(MultivariateDistribution):
                    def __init__(self,alpha):
                        self.alpha = alpha

                        a0 = sum(alpha)
                        b = (a0**2*(a0 + 1))
                        self._x = [a/a0 for a in alpha]
                        self._u = [a*(a0 - a)/b for a in alpha]
                        self._cov = [[ai*(a0 - ai)/b if i == j else -ai*aj/b
                                      for i,ai in enumerate(alpha)]
                                      for j,aj in enumerate(alpha)]

                        super().__init__(len(alpha))

                    def _simulate(self,n):
                        self.simdata = Distribution.random_state().dirichlet(self.alpha,n).T

                    def x(self):
                        return self._x

                    def u(self):
                        return self._u

                    @property
                    def cov(self):
                        return self._cov
        """
        self._elements = [MultivariateElement(self,i) for i in range(nd)]
        self._len = nd
        
    def __getitem__(self,i):
        return self._elements[i]
        
    def __iter__(self):
        return iter(self.elements)
        
    def __len__(self):
        return self._len
        
    def x(self):
        raise NotImplementedError()
        
    def u(self):
         raise NotImplementedError()
         
    @property
    def cov(self):
        raise NotImplementedError()
        
    def clear(self):
        self.simdata = None
         

class MultivariateElement(Distribution):
    def __init__(self,parent,index):
        """
        Represents one dimension of a multivariate distribution.  Do not create
        instances of this class directly.  Instances are created by the parent
        `MultivariateDistribution __init__` method.
        """
        self.parent = parent
        self.index = index
        
    def random(self,n=None):
        raise NotImplementedError('with a MuliVaraiteElement, data can only be generated with the simulate static method')
        
    def _simulate(self,n,ufrom):
        if ufrom is not None and self not in ufrom:
            self.simdata = np.full(n,self.x())
            return
        if self.parent not in Distribution._called:
            Distribution._called.append(self.parent)
            self.parent._simulate(n)
        self.simdata = self.parent.simdata[self.index]
        
    def cov(self,d):
        if not isinstance(d,MultivariateElement):
            return 0.0
        if d.parent is not self.parent:
            return 0.0
        return self.parent.cov[self.index][d.index]
        
    def clear(self):
        self.parent.simdata = None
        super().clear()
        
    def x(self):
        return self.parent.x()[self.index]
        
    def u(self):
        return self.parent.cov.diagonal()[self.index]

class NormalDist(Distribution):
    def __init__(self,x,s):
        """
        Normal distribution with mean `x` and standard deviation `s`.
        """
        if s < 0:
            raise ValueError('s < 0')
        self._x = x
        self._u = s
        
    def random(self,n=None):
        if self._u == 0:
            if n == None:
                return self._x
            else:
                return np.full(n,self._x)
        return Distribution.random_state().normal(self._x,self._u,n)
        
    def x(self):
        return self._x
        
    def u(self):
        return self._u
        
        
class TDist(Distribution):
    bayesian_default = False
    
    def __init__(self,x,s,dof):
        """
        Shifted and scaled Student's t distribution with `dof` degrees of freedom,
        mean `x`, and scale factor `s`.
        """
        if dof <= 0:
            raise ValueError('dof <= 0')
        self._x = x
        self._s = s
        self.dof = dof
        
    def random(self,n=None):
        if self._s == 0:
            if n == None:
                return self._x
            else:
                return np.full(n,self._x)
        return self._x + self._s*Distribution.random_state().standard_t(self.dof,n)
        
    def x(self):
        return self._x
        
    def u(self):
        return self._s

            
class MultiNormalDist(MultivariateDistribution):
    def __init__(self,mean,cov):
        """
        Multivariate normal distribution.  The parameter `mean` is a list of mean
        values for each dimension and `cov` is the variance-covariance matrix.
        """
        self.mean = np.asarray(mean)
        self._cov = np.asarray(cov)
        nd = len(self.mean)
        if cov.shape != (nd,nd):
            ValueError('cov.shape != (len(mean),len(mean))')
        super().__init__(nd)
            
    def _simulate(self,n):
        self.simdata = Distribution.random_state().multivariate_normal(self.mean,self.cov,n).T
        
    def x(self):
        return self.mean
        
    def u(self):
        return self.cov.diagonal()
        
    @property
    def cov(self):
        return self._cov
    
class MultiTElement(MultivariateElement):
    def __init__(self,parent,index,dof):
        """
        Represents one dimension of a `MultiTDist`.  Do not create
        instances of this class directly.  Instances are created by the parent
        `MultiTDist __init__` method.
        """
        self.parent = parent
        self.index = index
        self.dof = dof
        
class MultiTDist(MultivariateDistribution):
    def __init__(self,mean,cov,dof):
        """
        Multivariate shifted and scaled Student's t distribution.  The parameter 
        mean is a list of mean values for each dimension and cov is the 
        variance-covariance matrix.  The parameter dof is the number of
        degrees of freedom and must be scalar; all dimensions must have the same 
        number of degrees of freedom.
        """
        self.mean = np.asarray(mean)
        self._cov = np.asarray(cov)
        nd = len(self.mean)
        
        try:
            if len(dof) == nd:
                dof = np.asarray(dof)
                for d in dof:
                    if d != dof[0]:
                        raise ValueError('all dimensions must have the same dof')
            else:
                raise TypeError('dof must be a scalar or be the same length as mean')
        except TypeError:
            self.dof = np.array(nd*[dof])
        
        if cov.shape != (nd,nd):
            ValueError('cov.shape != (len(mean),len(mean))')
        self._elements = [MultiTElement(self,i,dof) for i in range(nd)]
        self._len = nd
    
    def _simulate(self,n):
        gamma = np.tile(np.random.gamma(self.dof[0]/2,2/self.dof[0],n),(len(self),1)).T
        Z = np.random.multivariate_normal(np.zeros(len(self)),self.cov,n)
        self.simdata = (self.mean + Z/np.sqrt(gamma)).T
        
    def x(self):
        return self.mean
        
    def u(self):
        return self.cov.diagonal()
        
    @property
    def cov(self):
        return self._cov
        
class UniformDist(Distribution):
    def __init__(self,center=None,half_width=None,lower_limit=None,upper_limit=None):
        """
        A uniform distribution.  Two and only two of the parameters `center`,
        `half_width`, `lower_limit` and `upper_limit` must be specified.
        """
        k = 0
        if center is not None:
            k += 1
            self.center = center
            if half_width is not None:
                self.lower_limit = center - half_width
                self.upper_limit = center + half_width
            elif lower_limit is not None:
                self.half_width = center - lower_limit
                self.upper_limit = center + self.half_width
            elif upper_limit is not None:
                self.half_width = upper_limit - center
                self.lower_limit = center - self.half_width
        if half_width is not None:
            k += 1
            if half_width <= 0:
                raise ValueError('half_width <= 0')
            self.half_width = half_width
            if lower_limit is not None:
                self.center = half_width + lower_limit
                self.upper_limit = self.center + half_width
            elif upper_limit is not None:
                self.center = upper_limit - half_width
                self.lower_limit = self.center - half_width 
        if lower_limit is not None:
            k += 1
            self.lower_limit = upper_limit
            if upper_limit is not None:
                if lower_limit >= upper_limit:
                    raise ValueError('lower_limit >= upper_limit')
                self.center = (upper_limit + lower_limit)/2
                self.half_width = upper_limit - lower_limit
        if upper_limit is not None:
            k += 1
            self.upper_limit = upper_limit
        if k != 2:
            raise ValueError('Two and only two of the parameters may be specified.')
        self._u = self.half_width/np.sqrt(3)
    
    def random(self,n=None):
        return Distribution.random_state().uniform(self.lower_limit,self.upper_limit,n)
    
    def x(self):
        return self.center
        
    def u(self):
        return self._u
    
class GammaDist(Distribution):
    def __init__(self,shape,scale):
        """
        Gamma distribution with the `shape` and `scale` parameters.
        """
        if shape <= 0:
            raise ValueError('shape < 0')
        if scale <= 0:
            raise ValueError('scale < 0')
        self.shape = shape
        self.scale = scale
        
    def random(self,n=None):
        return Distribution.random_state().gamma(self.shape,self.scale,n)
        
    def x(self):
        return self.shape*self.scale
        
    def u(self):
        return self.shape*np.sqrt(self.scale)
        
class LaplaceDist(Distribution):
    def __init__(self,x,scale):
        """
        Laplace distribution with location parameter `x` and `scale` parameter.
        """
        if scale <= 0:
            raise ValueError('scale < 0')
        self._x = x
        self.scale = scale
        
    def random(self,n=None):
        return Distribution.random_state().laplace(self.x(),self.scale,n)
        
    def x(self):
        return self._x
        
    def u(self):
        return self.scale*np.sqrt(2)
        
class TriangularDist(Distribution):
    def __init__(self,mode,half_width=None,left_width=None,right_width=None,lower_limit=None,upper_limit=None):
        """
        Triangular distribution.  `mode` is required, then, for a symmetric distribution
        specify `half_width`,  and for non-symmetric distributions specify two and only
        two of the parameters `left_width`, `right_width`, `lower_limit`, `upper_limit`.
        """
        self.mode = mode
        if half_width is not None:
            if half_width <= 0:
                raise ValueError('half_width <= 0')
            self.half_width = half_width
            self.left_width = half_width
            self.right_width = half_width
            if left_width is not None:
                raise ValueError('left_width and half_width both specified')
            if right_width is not None:
                raise ValueError('right_width and half_width both specified')
            if mode is not None:
                self.mode = mode
                if lower_limit is not None or upper_limit is not None:
                    raise ValueError('The distribution parameters are over specified.')
                self.lower_limit = mode-half_width
                self.upper_limit = mode+half_width
        else:
            if left_width is not None:
                if left_width <= 0:
                    raise ValueError('left_width <= 0')
                if lower_limit is not None:
                    raise ValueError('The distribution parameters are over specified.')
                self.lower_limit = mode-left_width
            else:
                if lower_limit is None:
                    raise ValueError('The distribution parameters are under specified.')
                if upper_limit is not None and upper_limit <= lower_limit:
                    raise ValueError('upper_limit <= lower_limit')
                self.left_width = mode-upper_limit
            if right_width is not None:
                if right_width <= 0:
                    raise ValueError('right_width <= 0')
                if upper_limit is not None:
                    raise ValueError('The distribution parameters are over specified.')
                self.upper_limit = mode-right_width
            else:
                if upper_limit is None:
                    raise ValueError('The distribution parameters are under specified.')
                self.right_width = mode-upper_limit
            
    def random(self,n=None):
        return Distribution.random_state().triangular(self.mode,self.lower_limit,self.upper_limit,n)
        
    def x(self):
        return self.mode
        
    def u(self):
        return np.sqrt((self.lower_limit**2+self.upper_limit**2)/6)
        
class ExponentialDist(Distribution):
    def __init__(self,scale=None,rate=None):
        """
        Exponential distribution:
        
        f(x;rate) = rate*exp(-rate*x)
        
        Specify either `scale` or `rate` (`scale` = 1/`rate`).
        """
        if scale is not None and rate is not None:
            raise ValueError('Both scale and rate may not both be specified.')
        if scale is not None:
            if scale <= 0:
                raise ValueError('scale <= 0')
            self.scale = scale
            self.rate = 1/scale
        if rate is not None:
            if rate <= 0:
                raise ValueError('rate <= 0')
            self.rate = rate
            self.scale = 1/rate
            
    def random(self,n=None):
        return Distribution.random_state().exponential(self.scale,n)
        
    def x(self):
        return self.scale
        
    def u(self):
        return self.scale
        
class PoissonDist(Distribution):
    def __init__(self,lam):
        """
        Poisson distribution with rate parameter `lam`.
        """
        if lam <= 0:
            raise ValueError('lam <= 0')
        if np.modf(lam)[1] != 0:
            raise ValueError('lam is not an integer')
        self.lam = lam
            
    def random(self,n=None):
        return Distribution.random_state().poisson(self.lam,n)
        
    def x(self):
        return self.lam
        
    def u(self):
        return np.sqrt(self.lam)
        
        
class BinomialDist(Distribution):
    def __init__(self,n,p):
        """
        Binomial distribution with number of trials `n` and success probability `p`.
        """
        if n <= 0:
            raise ValueError('n <= 0')
        if np.modf(n)[1] != 0:
            raise ValueError('n is not an integer')
        if p <= 0:
            raise ValueError('p <= 0')
        if np.modf(p)[1] != 0:
            raise ValueError('p is not an integer')
        self.n = n
        self.p = p
            
    def random(self,n=None):
        return Distribution.random_state().binomial(self.n,self.p,n)
        
    def x(self):
        return self.n*self.p
        
    def u(self):
        return np.sqrt(self.n*self.p(1-self.p))
        

class CurvlinearTrapDist(Distribution):
    def __init__(self,center=None,half_width=None,limit_half_range = None,
                 lower_limit=None,upper_limit=None):
        """
        Curvlinear trapezoidal distribution, `limit_half_range` is required.  Also
        either `center` and `half_width` or `lower_limit` and `upper_limit` are required.
        
        This is intended to represent a variable that follows a uniform distribution
        but where the upper and lower limits are not exactly known and may vary by
        up to the `limit_half_range` from the given lower and upper limit values.
        """
        if limit_half_range is None:
            raise ValueError('The limit_half_range must be specified.')
        self.limit_half_range = limit_half_range
        k = 0
        if center is not None:
            k += 1
            self.center = center
            if half_width is not None:
                self.lower_limit = center - half_width
                self.upper_limit = center + half_width
            elif lower_limit is not None:
                self.half_width = center - lower_limit
                self.upper_limit = center + self.half_width
            elif upper_limit is not None:
                self.half_width = upper_limit - center
                self.lower_limit = center - self.half_width
        if half_width is not None:
            k += 1
            if half_width <= 0:
                raise ValueError('half_width <= 0')
            self.half_width = half_width
            if lower_limit is not None:
                self.center = half_width + lower_limit
                self.upper_limit = self.center + half_width
            elif upper_limit is not None:
                self.center = upper_limit - half_width
                self.lower_limit = self.center - half_width 
        if lower_limit is not None:
            k += 1
            self.lower_limit = upper_limit
            if upper_limit is not None:
                if lower_limit >= upper_limit:
                    raise ValueError('lower_limit >= upper_limit')
                self.center = (upper_limit + lower_limit)/2
                self.half_width = upper_limit - lower_limit
        if upper_limit is not None:
            k += 1
            self.upper_limit = upper_limit
        if k != 2:
            raise ValueError('two and only two of the parameters may be specified')
    
    def random(self,n=None):
        r1 = Distribution.random_state().uniform(0,1,n)
        r2 = Distribution.random_state().uniform(0,1,n)
        sa = (self.lower_limt - self.limit_half_range) + 2*self.limit_half_range*r1
        sb = (self.lower_limit + self.upper_limit) - sa
        return sa+(sb - sa)*r2

    def x(self):
        return self.center
        
    def u(self):
        return np.sqrt(self.half_width**2/3 + self.limit_half_range**2/9)
        
class TrapezoidalDist(Distribution):
    def __init__(self,lower_limit,upper_limit,top_to_base_ratio):
        """
        Trapezoidal distribution
        """
        if top_to_base_ratio > 1:
            raise ValueError('top_to_base_ratio > 1')
        if lower_limit >= upper_limit:
            raise ValueError('lower_limit >= upper_limit')
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        
    def random(self,n=None):
        r1 = Distribution.random_state().uniform(0,1,n)
        r2 = Distribution.random_state().uniform(0,1,n)
        return (self.lower_limit + ((self.upper_limit - self.lower_limit)/2)*
                ((1+self.top_to_base_ratio)*r1 + (1-self.top_to_base_ratio)*r2))
    
    def x(self):
        return (self.upper_limit - self.lower_limit)/2
        
    def u(self):
        return (self.upper_limit - self.lower_limit)*np.sqrt((1+self.top_to_base_ratio**2)/24)
        
class ArcSinDist(Distribution):
    def __init__(self,center=None,half_width=None,lower_limit=None,upper_limit=None):
        """
        Arcsin distribution.  Specify either `center` and `half_width` or `lower_limit`
        and `upper_limit`.
        """
        k = 0
        if center is not None:
            k += 1
            self.center = center
            if half_width is not None:
                self.lower_limit = center - half_width
                self.upper_limit = center + half_width
            elif lower_limit is not None:
                self.half_width = center - lower_limit
                self.upper_limit = center + self.half_width
            elif upper_limit is not None:
                self.half_width = upper_limit - center
                self.lower_limit = center - self.half_width
        if half_width is not None:
            k += 1
            if half_width <= 0:
                raise ValueError('half_width <= 0')
            self.half_width = half_width
            if lower_limit is not None:
                self.center = half_width + lower_limit
                self.upper_limit = self.center + half_width
            elif upper_limit is not None:
                self.center = upper_limit - half_width
                self.lower_limit = self.center - half_width 
        if lower_limit is not None:
            k += 1
            self.lower_limit = upper_limit
            if upper_limit is not None:
                if lower_limit >= upper_limit:
                    raise ValueError('lower_limit >= upper_limit')
                self.center = (upper_limit + lower_limit)/2
                self.half_width = upper_limit - lower_limit
        if upper_limit is not None:
            k += 1
            self.upper_limit = upper_limit
        if k != 2:
            raise ValueError('two and only two of the parameters may be specified')
            
    def random(self,n=None):
        return (self.center + 
            self.half_width*np.sin(2*np.pi*Distribution.random_state().uniform(0,1,n)))
            
    def x(self):
        return self.center
        
    def u(self):
        return self.half_width/(2*np.sqrt(2))
        
class LogNormalDist(Distribution):
    def __init__(self,mu,sigma):
        """
        Log-normal distribution, `mu` and `sigma` are the mean and standard deviation
        of the logarithm of the random variable.
        """
        self.mu = mu
        self.sigma = sigma
            
    def random(self,n=None):
        return Distribution.random_state().lognormal(self.mu,self.sigma,n)
        
    def x(self):
        return np.exp(self.mu)
        
    def u(self):
       return self.sigma*np.exp(self.mu)
        
class WeibullDist(Distribution):
    def __init__(self,shape,scale):
        """
        Weibull distribution with `shape` and `scale` parameters.
        """
        from scipy.special import gamma
        self._gamma = gamma
        if shape < 0:
            raise ValueError('shape < 0')
        if scale <= 0:
            raise ValueError('scale <= 0')
        self.shape = shape
        self.scale = scale
            
    def random(self,n=None):
        return self.scale*Distribution.random_state().weibull(self.shape,n)
        
    def x(self):
        return self.scale*self._gamma(1 + 1/self.shape)
        
    def u(self):
        return self.scale*np.sqrt(self._gamma(1+2/self.shape) - self._gamma(1+1/self.shape)**2)
        
        
